# Copyright (C) 2024 qBraid
# Copyright (C) The Cirq Developers
#
# This file is part of the qBraid-SDK.
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. This specific file, adapted from Cirq, is dual-licensed under both the
# Apache License, Version 2.0, and the GPL v3. You may not use this file except in
# compliance with the applicable license. You may obtain a copy of the Apache License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file includes code adapted from Cirq (https://github.com/quantumlib/Cirq)
# with modifications by qBraid. The original copyright notice is included above.
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header

# pylint: disable=missing-function-docstring,too-many-lines

"""
Unit tests for Cirq Qasm Parser

"""

from typing import Callable

import cirq
import cirq.testing as ct
import numpy as np
import pytest
import sympy
from cirq import Circuit
from cirq.circuits.qasm_output import QasmUGate
from cirq.contrib.qasm_import import QasmException

import qbraid.transpiler.conversions.qasm2.cirq_custom as cirq_qasm_gates
from qbraid.transpiler.conversions.qasm2.cirq_qasm_parser import QasmParser


def test_format_header_circuit():
    parser = QasmParser()

    parsed_qasm = parser.parse("OPENQASM 2.0;")

    assert parsed_qasm.supportedFormat
    assert not parsed_qasm.qelib1Include
    ct.assert_same_circuits(parsed_qasm.circuit, Circuit())


def test_unsupported_format():
    qasm = "OPENQASM 2.1;"
    parser = QasmParser()

    with pytest.raises(QasmException, match="Unsupported.*2.1.*2.0.*supported.*"):
        parser.parse(qasm)


def test_format_header_with_quelibinc_circuit():
    qasm = """OPENQASM 2.0;
     include "qelib1.inc";
    """
    parser = QasmParser()

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include
    ct.assert_same_circuits(parsed_qasm.circuit, Circuit())


@pytest.mark.parametrize("qasm", ['include "qelib1.inc";', "", "qreg q[3];"])
def test_error_not_starting_with_format(qasm: str):
    parser = QasmParser()

    with pytest.raises(QasmException, match="Missing 'OPENQASM 2.0;' statement"):
        parser.parse(qasm)


def test_comments():
    parser = QasmParser()

    parsed_qasm = parser.parse(
        """
    //this is the format
    OPENQASM 2.0;
    // this is some other comment
    include "qelib1.inc";
    // and something at the end of the file
    // multiline
    """
    )

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include
    ct.assert_same_circuits(parsed_qasm.circuit, Circuit())


def test_multiple_qreg_declaration():
    qasm = """OPENQASM 2.0;
     include "qelib1.inc";
     qreg a_quantum_register [ 1337 ];
     qreg q[42];
    """
    parser = QasmParser()

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include
    ct.assert_same_circuits(parsed_qasm.circuit, Circuit())
    assert parsed_qasm.qregs == {"a_quantum_register": 1337, "q": 42}


@pytest.mark.parametrize(
    "qasm",
    [
        """OPENQASM 2.0;
           qreg q[2];
           creg q[3];
               """,
        """OPENQASM 2.0;
           creg q[2];
           qreg q[3];
               """,
    ],
)
def test_already_defined_error(qasm: str):
    parser = QasmParser()

    with pytest.raises(QasmException, match=r"q.*already defined.* line 3"):
        parser.parse(qasm)


@pytest.mark.parametrize(
    "qasm",
    [
        """OPENQASM 2.0;
           qreg q[0];
               """,
        """OPENQASM 2.0;
           creg q[0];
               """,
    ],
)
def test_zero_length_register(qasm: str):
    parser = QasmParser()

    with pytest.raises(QasmException, match=".* zero-length.*'q'.*line 2"):
        parser.parse(qasm)


def test_unexpected_end_of_file():
    qasm = """OPENQASM 2.0;
              include "qelib1.inc";
              creg
           """
    parser = QasmParser()

    with pytest.raises(QasmException, match="Unexpected end of file"):
        parser.parse(qasm)


def test_multiple_creg_declaration():
    qasm = """OPENQASM 2.0;
     include "qelib1.inc";
     creg a_classical_register [1337];
     qreg a_quantum_register [1337];
     creg c[42];
    """
    parser = QasmParser()

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include
    ct.assert_same_circuits(parsed_qasm.circuit, Circuit())
    assert parsed_qasm.qregs == {"a_quantum_register": 1337}
    assert parsed_qasm.cregs == {"a_classical_register": 1337, "c": 42}


def test_syntax_error():
    qasm = """OPENQASM 2.0;
         qreg q[2] bla;
         foobar q[0];
    """
    parser = QasmParser()

    with pytest.raises(QasmException, match=r"""Syntax error: 'bla'.*"""):
        parser.parse(qasm)


def test_CX_gate():
    qasm = """OPENQASM 2.0;
     qreg q1[2];
     qreg q2[2];
     CX q1[0], q1[1];
     CX q1, q2[0];
     CX q2, q1;      
    """
    parser = QasmParser()

    q1_0 = cirq.NamedQubit("q1_0")
    q1_1 = cirq.NamedQubit("q1_1")
    q2_0 = cirq.NamedQubit("q2_0")
    q2_1 = cirq.NamedQubit("q2_1")

    expected_circuit = Circuit()
    # CX q1[0], q1[1];
    expected_circuit.append(cirq.CNOT(q1_0, q1_1))
    # CX q1, q2[0];
    expected_circuit.append(cirq.CNOT(q1_0, q2_0))
    expected_circuit.append(cirq.CNOT(q1_1, q2_0))
    # CX q2, q1;
    expected_circuit.append(cirq.CNOT(q2_0, q1_0))
    expected_circuit.append(cirq.CNOT(q2_1, q1_1))

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert not parsed_qasm.qelib1Include

    ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q1": 2, "q2": 2}


@pytest.mark.skip(reason="Not yet supported")
def test_classical_control():
    qasm = """OPENQASM 2.0;
        qreg q[2];
        creg a[1];
        measure q[0] -> a[0];
        if (a==1) CX q[0],q[1];
    """
    parser = QasmParser()

    q_0 = cirq.NamedQubit("q_0")
    q_1 = cirq.NamedQubit("q_1")
    expected_circuit = cirq.Circuit(
        cirq.measure(q_0, key="a_0"),
        cirq.CNOT(q_0, q_1).with_classical_controls(sympy.Eq(sympy.Symbol("a_0"), 1)),
    )

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert not parsed_qasm.qelib1Include

    ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q": 2}

    # Note this cannot *exactly* round-trip because the way QASM and Cirq handle measurements
    # into classical registers is different. Cirq parses QASM classical registers into m_a_i for i
    # in 0..bit_count. Thus the generated key has an extra "_0" at the end.
    expected_generated_qasm = f"""// Generated from Cirq v{cirq.__version__}
OPENQASM 2.0;
include "qelib1.inc";
// Qubits: [q_0, q_1]
qreg q[2];
creg m_a_0[1];
measure q[0] -> m_a_0[0];
if (m_a_0==1) cx q[0],q[1];
"""
    assert cirq.qasm(parsed_qasm.circuit) == expected_generated_qasm


@pytest.mark.skip(reason="Not yet supported")
def test_classical_control_multi_bit():
    qasm = """OPENQASM 2.0;
        qreg q[2];
        creg a[2];
        measure q[0] -> a[0];
        measure q[0] -> a[1];
        if (a==1) CX q[0],q[1];
    """
    parser = QasmParser()

    q_0 = cirq.NamedQubit("q_0")
    q_1 = cirq.NamedQubit("q_1")

    # Since we split the measurement into two, we also need two conditions.
    # m_a==1 corresponds to m_a[0]==1, m_a[1]==0
    expected_circuit = cirq.Circuit(
        cirq.measure(q_0, key="a_0"),
        cirq.measure(q_0, key="a_1"),
        cirq.CNOT(q_0, q_1).with_classical_controls(
            sympy.Eq(sympy.Symbol("a_0"), 1), sympy.Eq(sympy.Symbol("a_1"), 0)
        ),
    )

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert not parsed_qasm.qelib1Include

    ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q": 2}

    # Note that this will *not* round-trip, but there's no good way around that due to the
    # difference in how Cirq and QASM do multi-bit measurements.
    with pytest.raises(ValueError, match="QASM does not support multiple conditions"):
        _ = cirq.qasm(parsed_qasm.circuit)


def test_CX_gate_not_enough_args():
    qasm = """OPENQASM 2.0;
     qreg q[2];
     CX q[0];
    """
    parser = QasmParser()

    with pytest.raises(QasmException, match=r"CX.*takes.*got.*1.*line 3"):
        parser.parse(qasm)


def test_CX_gate_mismatched_registers():
    qasm = """OPENQASM 2.0;
     qreg q1[2];
     qreg q2[3];
     CX q1, q2;
    """
    parser = QasmParser()

    with pytest.raises(QasmException, match=r"Non matching.*length \[2 3\].*line 4"):
        parser.parse(qasm)


def test_CX_gate_bounds():
    qasm = """OPENQASM 2.0;
     qreg q1[2];
     qreg q2[3];
     CX q1[4], q2[0];
    """
    parser = QasmParser()

    with pytest.raises(QasmException, match=r"Out of bounds.*4.*q1.*2.*line 4"):
        parser.parse(qasm)


def test_CX_gate_arg_overlap():
    qasm = """OPENQASM 2.0;
     qreg q1[2];
     qreg q2[3];
     CX q1[1], q1[1];
    """
    parser = QasmParser()

    with pytest.raises(QasmException, match=r"Overlapping.*at line 4"):
        parser.parse(qasm)


def test_U_gate():
    qasm = """
     OPENQASM 2.0;
     qreg q[2];
     U(pi, 2.3, 3) q[0];
     U(+3.14, -pi, (8)) q;
    """
    parser = QasmParser()

    q0 = cirq.NamedQubit("q_0")
    q1 = cirq.NamedQubit("q_1")

    expected_circuit = Circuit()
    expected_circuit.append(
        cirq.Moment(
            [
                QasmUGate(1.0, 2.3 / np.pi, 3 / np.pi)(q0),
                QasmUGate(3.14 / np.pi, -1.0, 8 / np.pi)(q1),
            ]
        )
    )

    expected_circuit.append(cirq.Moment([QasmUGate(3.14 / np.pi, -1.0, 8 / np.pi)(q0)]))

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert not parsed_qasm.qelib1Include

    ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q": 2}


def test_U_angles():
    qasm = """
    OPENQASM 2.0;
    qreg q[1];
    U(pi/2,0,pi) q[0];
    """

    c = QasmParser().parse(qasm).circuit
    cirq.testing.assert_allclose_up_to_global_phase(
        cirq.unitary(c), cirq.unitary(cirq.H), atol=1e-7
    )


def test_U_gate_zero_params_error():
    qasm = """OPENQASM 2.0;
     qreg q[2];     
     U q[1];"""

    parser = QasmParser()

    with pytest.raises(QasmException, match=r"U takes 3.*got.*0.*line 3"):
        parser.parse(qasm)


def test_U_gate_too_much_params_error():
    qasm = """OPENQASM 2.0;
     qreg q[2];     
     U(pi, pi, pi, pi) q[1];"""

    parser = QasmParser()

    with pytest.raises(QasmException, match=r"U takes 3.*got.*4.*line 3"):
        parser.parse(qasm)


@pytest.mark.parametrize(
    "expr",
    [
        ".333 + 4",
        "1.0 * 2",
        "0.1 ^ pi",
        "0.1 / pi",
        "2.0e-05 ^ (1/2)",
        "1.2E+05 * (3 + 2)",
        "123123.2132312 * cos(pi)",
        "123123.2132312 * sin(2 * pi)",
        "3 - 4 * 2",  # precedence of *
        "3 * 4 + 2",  # precedence of *
        "3 * 4 ^ 2",  # precedence of ^
        "3 - 4 ^ 2",  # precedence of ^
        "3^2^(-2)",  # right associativity of ^
        "(-1) * pi",
        "(+1) * pi",
        "-3 * 5 + 2",
        "(+4 * (-3) ^ 5 - 2)",
        "tan(123123.2132312)",
        "ln(pi)",
        "exp(2*pi)",
        "sqrt(4)",
        "acos(1)",
        "atan(0.2)",
    ],
)
def test_expressions(expr: str):
    qasm = f"""OPENQASM 2.0;
     qreg q[1];
     U({expr}, 2 * pi, pi / 2.0) q[0];
    """

    parser = QasmParser()

    q0 = cirq.NamedQubit("q_0")

    expected_circuit = Circuit()
    expected_circuit.append(QasmUGate(float(sympy.sympify(expr)) / np.pi, 2.0, 1 / 2.0)(q0))

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert not parsed_qasm.qelib1Include

    ct.assert_allclose_up_to_global_phase(
        cirq.unitary(parsed_qasm.circuit), cirq.unitary(expected_circuit), atol=1e-10
    )
    assert parsed_qasm.qregs == {"q": 1}


def test_unknown_function():
    qasm = """OPENQASM 2.0;
     qreg q[1];
     U(nonexistent(3), 2 * pi, pi / 3.0) q[0];
    """
    parser = QasmParser()

    with pytest.raises(QasmException, match=r".*not recognized.*'nonexistent'.*line 3"):
        parser.parse(qasm)


rotation_gates = [("rx", cirq.rx), ("ry", cirq.ry), ("rz", cirq.rz)]


single_qubit_gates = [
    ("x", cirq.X),
    ("y", cirq.Y),
    ("z", cirq.Z),
    ("h", cirq.H),
    ("s", cirq.S),
    ("t", cirq.T),
    ("sdg", cirq.S**-1),
    ("tdg", cirq.T**-1),
    ("sx", cirq.XPowGate(exponent=0.5)),
]


@pytest.mark.parametrize("qasm_gate,cirq_gate", rotation_gates)
def test_rotation_gates(qasm_gate: str, cirq_gate: Callable[[float], cirq.Gate]):
    qasm = f"""OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[2];
     {qasm_gate}(pi/2) q[0];
     {qasm_gate}(pi) q;
    """

    parser = QasmParser()

    q0 = cirq.NamedQubit("q_0")
    q1 = cirq.NamedQubit("q_1")

    expected_circuit = Circuit()
    expected_circuit.append(cirq.Moment([cirq_gate(np.pi / 2).on(q0), cirq_gate(np.pi).on(q1)]))
    expected_circuit.append(cirq.Moment([cirq_gate(np.pi).on(q0)]))

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include

    ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q": 2}


@pytest.mark.parametrize("qasm_gate", [g[0] for g in rotation_gates])
def test_rotation_gates_wrong_number_of_args(qasm_gate: str):
    qasm = f"""
     OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[2];
     {qasm_gate}(pi) q[0], q[1];
    """

    parser = QasmParser()

    with pytest.raises(QasmException, match=rf".*{qasm_gate}.* takes 1.*got.*2.*line 5"):
        parser.parse(qasm)


@pytest.mark.parametrize("qasm_gate", [g[0] for g in rotation_gates])
def test_rotation_gates_zero_params_error(qasm_gate: str):
    qasm = f"""OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[2];
     {qasm_gate} q[1];
    """

    parser = QasmParser()

    with pytest.raises(QasmException, match=rf".*{qasm_gate}.* takes 1.*got.*0.*line 4"):
        parser.parse(qasm)


def test_qelib_gate_without_include_statement():
    qasm = """OPENQASM 2.0;
         qreg q[2];
         x q[0];
    """
    parser = QasmParser()

    with pytest.raises(QasmException, match=r"""Unknown gate "x".* line 3.*forget.*\?"""):
        parser.parse(qasm)


def test_undefined_register_from_qubit_arg():
    qasm = """OPENQASM 2.0;
            qreg q[2];
            CX q[0], q2[1];
       """
    parser = QasmParser()

    with pytest.raises(QasmException, match=r"""Undefined.*register.*q2.*"""):
        parser.parse(qasm)


def test_undefined_register_from_register_arg():
    qasm = """OPENQASM 2.0;
            qreg q[2];
            qreg q2[2];
            CX q1, q2;
       """
    parser = QasmParser()

    with pytest.raises(QasmException, match=r"""Undefined.*register.*q.*"""):
        parser.parse(qasm)


def test_measure_individual_bits():
    qasm = """
         OPENQASM 2.0;
         include "qelib1.inc";
         qreg q1[2];
         creg c1[2];                        
         measure q1[0] -> c1[0];
         measure q1[1] -> c1[1];
    """
    parser = QasmParser()

    q1_0 = cirq.NamedQubit("q1_0")
    q1_1 = cirq.NamedQubit("q1_1")

    expected_circuit = Circuit()

    expected_circuit.append(cirq.MeasurementGate(num_qubits=1, key="c1_0").on(q1_0))
    expected_circuit.append(cirq.MeasurementGate(num_qubits=1, key="c1_1").on(q1_1))

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include

    ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q1": 2}
    assert parsed_qasm.cregs == {"c1": 2}


def test_measure_registers():
    qasm = """OPENQASM 2.0;
         include "qelib1.inc";
         qreg q1[3];
         creg c1[3];                        
         measure q1 -> c1;       
    """
    parser = QasmParser()

    q1_0 = cirq.NamedQubit("q1_0")
    q1_1 = cirq.NamedQubit("q1_1")
    q1_2 = cirq.NamedQubit("q1_2")

    expected_circuit = Circuit()

    expected_circuit.append(cirq.MeasurementGate(num_qubits=1, key="c1_0").on(q1_0))
    expected_circuit.append(cirq.MeasurementGate(num_qubits=1, key="c1_1").on(q1_1))
    expected_circuit.append(cirq.MeasurementGate(num_qubits=1, key="c1_2").on(q1_2))

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include

    ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q1": 3}
    assert parsed_qasm.cregs == {"c1": 3}


def test_measure_mismatched_register_size():
    qasm = """OPENQASM 2.0;
         include "qelib1.inc";       
         qreg q1[2];
         creg c1[3];                        
         measure q1 -> c1;       
    """

    parser = QasmParser()

    with pytest.raises(QasmException, match=r""".*mismatched .* 2 -> 3.*line 5"""):
        parser.parse(qasm)


def test_measure_to_quantum_register():
    qasm = """OPENQASM 2.0;
         include "qelib1.inc";       
         qreg q1[3];
         qreg q2[3];
         creg c1[3];                        
         measure q2 -> q1;       
    """

    parser = QasmParser()

    with pytest.raises(QasmException, match=r"""Undefined classical register.*q1.*line 6"""):
        parser.parse(qasm)


def test_measure_undefined_classical_bit():
    qasm = """OPENQASM 2.0;
         include "qelib1.inc";       
         qreg q1[3];    
         creg c1[3];                        
         measure q1[1] -> c2[1];       
    """

    parser = QasmParser()

    with pytest.raises(QasmException, match=r"""Undefined classical register.*c2.*line 5"""):
        parser.parse(qasm)


def test_measure_from_classical_register():
    qasm = """OPENQASM 2.0;
         include "qelib1.inc";       
         qreg q1[2];
         creg c1[3];                        
         creg c2[3];                        
         measure c1 -> c2;       
    """

    parser = QasmParser()

    with pytest.raises(QasmException, match=r"""Undefined quantum register.*c1.*line 6"""):
        parser.parse(qasm)


def test_measurement_bounds():
    qasm = """OPENQASM 2.0;
     qreg q1[3];
     creg c1[3];                        
     measure q1[0] -> c1[4];  
    """
    parser = QasmParser()

    with pytest.raises(QasmException, match=r"Out of bounds bit.*4.*c1.*size 3.*line 4"):
        parser.parse(qasm)


def test_u1_gate():
    qasm = """
     OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[1];
     u1(pi / 3.0) q[0];
    """
    parser = QasmParser()

    q0 = cirq.NamedQubit("q_0")

    expected_circuit = Circuit()
    expected_circuit.append(cirq.ops.ZPowGate(exponent=1.0 / 3.0)(q0))

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include

    ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q": 1}


def test_u2_gate():
    qasm = """
     OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[1];
     u2(2 * pi, pi / 3.0) q[0];
    """
    parser = QasmParser()

    q0 = cirq.NamedQubit("q_0")

    expected_circuit = Circuit()
    expected_circuit.append(cirq_qasm_gates.U2Gate(2.0 * np.pi, np.pi / 3.0)(q0))

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include

    qasm_unitary = parsed_qasm.circuit.unitary()
    expected_unitary = expected_circuit.unitary()
    assert np.allclose(qasm_unitary, expected_unitary)

    # ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q": 1}


def test_id_gate():
    qasm = """
     OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[2];
     id q;           
    """
    parser = QasmParser()

    q0 = cirq.NamedQubit("q_0")
    q1 = cirq.NamedQubit("q_1")

    expected_circuit = Circuit()
    expected_circuit.append(cirq.IdentityGate(num_qubits=1)(q0))
    expected_circuit.append(cirq.IdentityGate(num_qubits=1)(q1))

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include

    ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q": 2}


def test_u3_gate():
    qasm = """
     OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[2];
     u3(pi, 2.3, 3) q[0];
     u3(+3.14, -pi, (8)) q;
    """
    parser = QasmParser()

    q0 = cirq.NamedQubit("q_0")
    q1 = cirq.NamedQubit("q_1")

    expected_circuit = Circuit()
    expected_circuit.append(
        cirq.Moment(
            [
                cirq_qasm_gates.U3Gate(np.pi, 2.3, 3.0)(q0),
                cirq_qasm_gates.U3Gate(3.14, -np.pi, 8.0)(q1),
            ]
        )
    )

    expected_circuit.append(cirq.Moment([cirq_qasm_gates.U3Gate(3.14, -np.pi, 8.0)(q0)]))

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include

    qasm_unitary = parsed_qasm.circuit.unitary()
    expected_unitary = expected_circuit.unitary()
    assert np.allclose(qasm_unitary, expected_unitary)

    # ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q": 2}


def test_r_gate():
    qasm = """
     OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[1];
     r(pi, pi / 2.0) q[0];    
    """
    parser = QasmParser()

    q0 = cirq.NamedQubit("q_0")

    expected_circuit = Circuit()
    expected_circuit.append(QasmUGate(1.0, 0.0, 0.0)(q0))

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include

    ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q": 1}


@pytest.mark.parametrize(
    "qasm_gate",
    ["id", "u2", "u3", "r"] + [g[0] for g in rotation_gates] + [g[0] for g in single_qubit_gates],
)
def test_standard_single_qubit_gates_wrong_number_of_args(qasm_gate):
    qasm = f"""
     OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[2];
     {qasm_gate} q[0], q[1];
    """

    parser = QasmParser()

    with pytest.raises(QasmException, match=r".* takes 1.*got.*2.*line 5"):
        parser.parse(qasm)


@pytest.mark.parametrize(
    ["qasm_gate", "num_params"],
    [["id", 0], ["u2", 2], ["u3", 3], ["rx", 1], ["ry", 1], ["rz", 1], ["r", 2]]
    + [[g[0], 0] for g in single_qubit_gates],
)
def test_standard_gates_wrong_params_error(qasm_gate: str, num_params: int):
    qasm = f"""OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[2];
     {qasm_gate}(pi, 2*pi, 3*pi, 4*pi, 5*pi) q[1];
    """

    parser = QasmParser()

    with pytest.raises(QasmException, match=rf".*{qasm_gate}.* takes {num_params}.*got.*5.*line 4"):
        parser.parse(qasm)

    if num_params == 0:
        return

    qasm = f"""OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[2];
     {qasm_gate} q[1];
    """

    parser = QasmParser()

    with pytest.raises(QasmException, match=rf".*{qasm_gate}.* takes {num_params}.*got.*0.*line 4"):
        parser.parse(qasm)


two_qubit_gates = [
    ("cx", cirq.CNOT),
    ("CX", cirq.CNOT),
    ("cz", cirq.CZ),
    ("cy", cirq.ControlledGate(cirq.Y)),
    ("swap", cirq.SWAP),
    ("ch", cirq.ControlledGate(cirq.H)),
]


@pytest.mark.parametrize("qasm_gate,cirq_gate", two_qubit_gates)
def test_two_qubit_gates(qasm_gate: str, cirq_gate: cirq.testing.TwoQubitGate):
    qasm = f"""
     OPENQASM 2.0;
     include "qelib1.inc";
     qreg q1[2];
     qreg q2[2];
     {qasm_gate} q1[0], q1[1];
     {qasm_gate} q1, q2[0];
     {qasm_gate} q2, q1;
    """
    parser = QasmParser()

    q1_0 = cirq.NamedQubit("q1_0")
    q1_1 = cirq.NamedQubit("q1_1")
    q2_0 = cirq.NamedQubit("q2_0")
    q2_1 = cirq.NamedQubit("q2_1")

    expected_circuit = Circuit()
    # CX q1[0], q1[1];
    expected_circuit.append(cirq_gate(q1_0, q1_1))
    # CX q1, q2[0];
    expected_circuit.append(cirq_gate(q1_0, q2_0))
    expected_circuit.append(cirq_gate(q1_1, q2_0))
    # CX q2, q1;
    expected_circuit.append(cirq_gate(q2_0, q1_0))
    expected_circuit.append(cirq_gate(q2_1, q1_1))

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include

    ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q1": 2, "q2": 2}


@pytest.mark.parametrize("qasm_gate", [g[0] for g in two_qubit_gates])
def test_two_qubit_gates_not_enough_args(qasm_gate: str):
    qasm = f"""
     OPENQASM 2.0;    
     include "qelib1.inc";             
     qreg q[2];
     {qasm_gate} q[0];
    """

    parser = QasmParser()

    with pytest.raises(QasmException, match=rf".*{qasm_gate}.* takes 2 arg\(s\).*got.*1.*line 5"):
        parser.parse(qasm)


@pytest.mark.parametrize("qasm_gate", [g[0] for g in two_qubit_gates])
def test_two_qubit_gates_with_too_much_parameters(qasm_gate: str):
    qasm = f"""
     OPENQASM 2.0;    
     include "qelib1.inc";             
     qreg q[2];
     {qasm_gate}(pi) q[0],q[1];
    """

    parser = QasmParser()

    with pytest.raises(
        QasmException, match=rf".*{qasm_gate}.* takes 0 parameter\(s\).*got.*1.*line 5"
    ):
        parser.parse(qasm)


three_qubit_gates = [("ccx", cirq.TOFFOLI), ("cswap", cirq.CSWAP)]


@pytest.mark.parametrize("qasm_gate,cirq_gate", three_qubit_gates)
def test_three_qubit_gates(qasm_gate: str, cirq_gate: cirq.testing.TwoQubitGate):
    qasm = f"""
     OPENQASM 2.0;
     include "qelib1.inc";
     qreg q1[2];
     qreg q2[2];
     qreg q3[2];
     {qasm_gate} q1[0], q1[1], q2[0];
     {qasm_gate} q1, q2[0], q3[0];
     {qasm_gate} q1, q2, q3;
    """
    parser = QasmParser()

    q1_0 = cirq.NamedQubit("q1_0")
    q1_1 = cirq.NamedQubit("q1_1")
    q2_0 = cirq.NamedQubit("q2_0")
    q2_1 = cirq.NamedQubit("q2_1")
    q3_0 = cirq.NamedQubit("q3_0")
    q3_1 = cirq.NamedQubit("q3_1")

    expected_circuit = Circuit()

    expected_circuit.append(cirq_gate(q1_0, q1_1, q2_0))

    expected_circuit.append(cirq_gate(q1_0, q2_0, q3_0))
    expected_circuit.append(cirq_gate(q1_1, q2_0, q3_0))

    expected_circuit.append(cirq_gate(q1_0, q2_0, q3_0))
    expected_circuit.append(cirq_gate(q1_1, q2_1, q3_1))

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include

    ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q1": 2, "q2": 2, "q3": 2}


@pytest.mark.parametrize("qasm_gate", [g[0] for g in three_qubit_gates])
def test_three_qubit_gates_not_enough_args(qasm_gate: str):
    qasm = f"""OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[2];
     {qasm_gate} q[0];
    """

    parser = QasmParser()

    with pytest.raises(
        QasmException, match=rf""".*{qasm_gate}.* takes 3 arg\(s\).*got.*1.*line 4"""
    ):
        parser.parse(qasm)


@pytest.mark.parametrize("qasm_gate", [g[0] for g in three_qubit_gates])
def test_three_qubit_gates_with_too_much_parameters(qasm_gate: str):
    qasm = f"""OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[3];
     {qasm_gate}(pi) q[0],q[1],q[2];
    """

    parser = QasmParser()

    with pytest.raises(QasmException, match=rf""".*{qasm_gate}.*parameter.*line 4.*"""):
        parser.parse(qasm)


@pytest.mark.parametrize("qasm_gate,cirq_gate", single_qubit_gates)
def test_single_qubit_gates(qasm_gate: str, cirq_gate: cirq.Gate):
    qasm = f"""OPENQASM 2.0;
     include "qelib1.inc";
     qreg q[2];
     {qasm_gate} q[0];
     {qasm_gate} q;
    """

    parser = QasmParser()

    q0 = cirq.NamedQubit("q_0")
    q1 = cirq.NamedQubit("q_1")

    expected_circuit = Circuit([cirq_gate.on(q0), cirq_gate.on(q0), cirq_gate.on(q1)])

    parsed_qasm = parser.parse(qasm)

    assert parsed_qasm.supportedFormat
    assert parsed_qasm.qelib1Include

    ct.assert_same_circuits(parsed_qasm.circuit, expected_circuit)
    assert parsed_qasm.qregs == {"q": 2}



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name,unused-argument

"""
Unit tests for runtime display functions.

"""
from unittest.mock import patch

from qbraid.runtime._display import display_jobs_from_data


def test_empty_table():
    """Test displaying an empty table."""
    data = {}
    with patch("builtins.print") as mock_print:
        display_jobs_from_data(data)
        actual_print_args = " ".join([str(arg[0]) for arg in mock_print.call_args_list])
        assert "No jobs found matching criteria." in actual_print_args


def test_nonempty_table():
    """Test displaying a nonempty table."""
    data = [("1", "2", "3")]
    with patch("builtins.print") as mock_print:
        display_jobs_from_data(data)
        actual_print_args = " ".join([str(arg[0]) for arg in mock_print.call_args_list])
        assert "Displaying 1 job." in actual_print_args



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining QuboProgram Class

"""
from __future__ import annotations

from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.typer import QuboCoefficientsDict

from ._model import AnnealingProgram, QuboProblem


class QuboProgram(AnnealingProgram):
    """AnnealingProblem subclass that accepts a qbraid.typer.QuboCoefficientsDict."""

    def __init__(self, program: QuboCoefficientsDict):
        super().__init__(program)
        if not isinstance(program, QuboCoefficientsDict):
            raise ProgramTypeError(
                message=f"Expected '{QuboCoefficientsDict}' object, got '{type(program)}'."
            )

    def to_problem(self) -> QuboProblem:
        """Converts the QuboCoefficientsDict a Problem instance."""
        return QuboProblem(self.program)



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing AutoQASM programs used for testing
"""
import autoqasm as aq
import numpy as np
from autoqasm.instructions import (
    cnot,
    cphaseshift,
    h,
    iswap,
    phaseshift,
    rx,
    ry,
    rz,
    s,
    si,
    swap,
    t,
    ti,
    v,
    vi,
    x,
    y,
    z,
)

AutoQASMType = aq.program.program.Program


@aq.main
def autoqasm_bell_circuit():
    h(0)
    cnot(0, 1)


def autoqasm_bell() -> AutoQASMType:
    """Returns AutoQASM bell circuit."""
    return autoqasm_bell_circuit.build()


@aq.main
def autoqasm_shared15_circuit():
    for i in range(4):
        h(i)
    x(0)
    x(1)
    y(2)
    z(3)
    s(0)
    si(1)
    t(2)
    ti(3)
    rx(0, np.pi / 4)
    ry(1, np.pi / 2)
    rz(2, 3 * np.pi / 4)
    phaseshift(3, np.pi / 8)
    v(0)
    vi(1)
    iswap(2, 3)
    swap(0, 2)
    swap(1, 3)
    cnot(0, 1)
    cphaseshift(2, 3, np.pi / 4)


def autoqasm_shared15() -> AutoQASMType:
    """Returns AutoQASM `Circuit` for qBraid `TestSharedGates`."""
    return autoqasm_shared15_circuit.build()



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining Azure Provider class for retrieving all Azure backends.

"""
from __future__ import annotations

import json
import os
import warnings
from typing import TYPE_CHECKING, Optional

from azure.identity import ClientSecretCredential
from azure.quantum import Workspace
from azure.quantum.target import Target
from qbraid_core._import import LazyLoader

from qbraid._caching import cached_method
from qbraid.programs import ExperimentType, IonQDict, ProgramSpec
from qbraid.runtime.exceptions import ResourceNotFoundError
from qbraid.runtime.profile import TargetProfile
from qbraid.runtime.provider import QuantumProvider

from .device import AzureQuantumDevice
from .io_format import InputDataFormat

if TYPE_CHECKING:
    from pulser.sequence import Sequence

pyquil = LazyLoader("pyquil", globals(), "pyquil")
pyqir = LazyLoader("pyqir", globals(), "pyqir")
pulser = LazyLoader("pulser", globals(), "pulser")

DEVICE_NUM_QUBITS = {
    "ionq.simulator": 29,
    "ionq.qpu.aria-1": 25,
    "ionq.qpu.aria-2": 25,
    "ionq.qpu.forte": 32,
    "quantinuum.sim.h1-1sc": 20,
    "quantinuum.sim.h2-1sc": 56,
    "quantinuum.sim.h2-2sc": 56,
    "quantinuum.sim.h1-1e": 20,
    "quantinuum.sim.h2-1e": 32,
    "quantinuum.sim.h2-2e": 32,
    "quantinuum.qpu.h1-1": 20,
    "quantinuum.qpu.h2-1": 56,
    "quantinuum.qpu.h2-2": 56,
    "rigetti.sim.qvm": None,
    "rigetti.qpu.ankaa-3": 84,
    "pasqal.sim.emu-tn": 100,
    "pasqal.qpu.fresnel": 100,
}


def serialize_pulser_input(seq: Sequence) -> str:
    """Convert a Pulser sequence to a JSON string."""
    input_data = {}
    input_data["sequence_builder"] = json.loads(seq.to_abstract_repr())
    to_send = json.dumps(input_data)
    return to_send


class AzureQuantumProvider(QuantumProvider):
    """
    Manages interactions with Azure Quantum services, encapsulating API calls,
    handling Azure Storage, and managing sessions.

    Attributes:
        workspace (Workspace): The configured Azure Quantum workspace.
    """

    def __init__(
        self,
        workspace: Optional[Workspace] = None,
        credential: Optional[ClientSecretCredential] = None,
    ):
        """
        Initializes an AzureQuantumProvider instance with a specified Workspace. It sets the
        credential for the workspace if it is not already set and a credential is provided.

        Args:
            workspace (Workspace, optional): An Azure Quantum Workspace object. If not provided,
                will be initialized with the provided credential or from environment variables.
            credential (ClientSecretCredential, optional): Optional credential to be used
                if the workspace lacks one.
        """
        if not workspace:
            connection_str = os.getenv("AZURE_QUANTUM_CONNECTION_STRING")
            workspace = (
                Workspace.from_connection_string(connection_str) if connection_str else Workspace()
            )

        if not workspace.credential:
            if credential:
                workspace.credential = credential
            else:
                warnings.warn(
                    "No credential provided; interactive authentication via a "
                    "web browser may be required. To avoid interactive authentication, "
                    "provide a ClientSecretCredential."
                )
        workspace.append_user_agent("qbraid")
        self._workspace = workspace

    @property
    def workspace(self) -> Workspace:
        """Get the Azure Quantum workspace."""
        return self._workspace

    @staticmethod
    def _get_program_spec(input_data_format: InputDataFormat) -> Optional[ProgramSpec]:
        """Get the program specification based on the input data format."""
        try:
            if input_data_format == InputDataFormat.MICROSOFT:
                return ProgramSpec(
                    pyqir.Module, alias="pyqir", serialize=lambda module: module.bitcode
                )
            if input_data_format == InputDataFormat.IONQ:
                return ProgramSpec(IonQDict, alias="ionq", serialize=lambda ionq_dict: ionq_dict)
            if input_data_format == InputDataFormat.QUANTINUUM:
                return ProgramSpec(str, alias="qasm2", serialize=lambda qasm: qasm)
            if input_data_format == InputDataFormat.RIGETTI:
                return ProgramSpec(
                    pyquil.Program, alias="pyquil", serialize=lambda program: program.out()
                )
            if input_data_format == InputDataFormat.PASQAL:
                return ProgramSpec(
                    pulser.sequence.Sequence, alias="pulser", serialize=serialize_pulser_input
                )
        except ModuleNotFoundError as err:  # pragma: no cover
            warnings.warn(
                f"The default runtime configuration for device using input data format "
                f"'{input_data_format.value}' requires package '{err.name}', "
                "which is not installed.",
                RuntimeWarning,
            )

        return None  # pragma: no cover

    @staticmethod
    def _get_experiment_type(input_data_format: InputDataFormat) -> ExperimentType:
        """Get the experiment type based on the input data format."""
        if input_data_format == InputDataFormat.PASQAL:
            return ExperimentType.AHS
        return ExperimentType.GATE_MODEL

    def _build_profile(self, target: Target) -> TargetProfile:
        """Builds a profile for an Azure device.

        Args:
            target (Target): The Target object to build profile from.

        Returns:
            TargetProfile: The completed profile.
        """
        device_id = target.name
        provider_name = target.provider_id
        simulator = device_id.split(".", 2)[1] != "qpu"
        capability = target.capability
        input_data_format = target.input_data_format
        output_data_format = target.output_data_format
        content_type = target.content_type

        num_qubits = DEVICE_NUM_QUBITS.get(device_id)

        try:
            input_data_format_enum = InputDataFormat(input_data_format)
        except ValueError:
            warnings.warn(f"Unrecognized input data format: {input_data_format}")
            experiment_type = None
            program_spec = None
        else:
            experiment_type = self._get_experiment_type(input_data_format_enum)
            program_spec = self._get_program_spec(input_data_format_enum)

        return TargetProfile(
            device_id=device_id,
            simulator=simulator,
            provider_name=provider_name,
            num_qubits=num_qubits,
            capability=capability,
            input_data_format=input_data_format,
            output_data_format=output_data_format,
            content_type=content_type,
            experiment_type=experiment_type,
            program_spec=program_spec,
        )

    @cached_method
    def get_devices(self, **kwargs) -> list[AzureQuantumDevice]:
        """Get all Azure Quantum devices.

        Args:
            **kwargs: Filters for the devices to retrieve.

        Returns:
            list[AzureQuantumDevice]: The Azure Quantum devices.

        """
        targets = self.workspace.get_targets(**kwargs)
        if isinstance(targets, Target):
            targets = [targets]

        if len(targets) == 0:
            if len(kwargs) > 0:
                raise ValueError("No devices found with the specified filters.")
            raise ResourceNotFoundError("No devices found.")

        return [
            AzureQuantumDevice(self._build_profile(target), self.workspace) for target in targets
        ]

    @cached_method
    def get_device(self, device_id: str) -> AzureQuantumDevice:
        """Get a specific Azure Quantum device.

        Args:
            device_id (str): The ID of the device to retrieve.

        Returns:
            AzureQuantumDevice: The Azure Quantum device.

        """
        target = self.workspace.get_targets(name=device_id)
        if not target:
            raise ValueError(f"Device {device_id} not found.")
        return AzureQuantumDevice(self._build_profile(target), self.workspace)

    def __hash__(self):
        if not hasattr(self, "_hash"):
            object.__setattr__(
                self, "_hash", hash((self._workspace.credential, self._workspace.user_agent))
            )
        return self._hash  # pylint: disable=no-member



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=expression-not-assigned
"""
Unit tests for OpenQASM2 to pyqpanda3 transpilation.

"""

from __future__ import annotations

import numpy as np
import pytest

from qbraid import transpile
from qbraid.interface import circuits_allclose

pyqpanda3 = pytest.importorskip("pyqpanda3")


def test_openqasm2_to_pyqpanda3():
    """Test converting an OpenQASM2 program to a pyqpanda3 QProg"""

    qasm2_str_in = """
    OPENQASM 2.0;
    include "stdgates.inc";
    // comment
    qubit[1] q;

    h q[0];
    x q[0];
    y q[0];
    """
    qprog = transpile(qasm2_str_in, target="pyqpanda3")

    expected = pyqpanda3.core.QProg()
    expected << pyqpanda3.core.H(0) << pyqpanda3.core.X(0) << pyqpanda3.core.Y(0)

    # pylint thinks .ndarray() has an extra argument, so disabling for the test
    # pylint: disable=no-value-for-parameter
    u_obs = pyqpanda3.quantum_info.Unitary(qprog.to_circuit()).ndarray()
    u_expected = pyqpanda3.quantum_info.Unitary(expected.to_circuit()).ndarray()
    assert np.allclose(u_obs, u_expected)


# Skip test as pyqpanda3 generates qasm with a creg of size 0 even
# though there are no classical bits in the circuit. pyqasm, which parses
# this output for circuits_allclose, errors that creg size must be > 0.
@pytest.mark.skip
def test_pyqpanda3_to_openqasm2():
    """Test converting a pyqpanda3 QProg to OpenQASM2"""

    prog = pyqpanda3.core.QProg()
    prog << pyqpanda3.core.RX(0, 3.1415926) << pyqpanda3.core.CNOT(0, 2)

    expected = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[3];
    rx(3.14159260) q[0];
    cx q[0],q[2];
    """

    res = transpile(prog, target="qasm2")
    assert circuits_allclose(res, expected)


# Until the above is supported, compare the QASM strings directly
def test_pyqpanda3_to_openqasm2_str_cmp():
    """Test converting a pyqpanda3 QProg to OpenQASM2"""

    prog = pyqpanda3.core.QProg()
    prog << pyqpanda3.core.RX(0, 3.1415926) << pyqpanda3.core.CNOT(0, 2)

    expected = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[3];
    creg c[0];
    rx(3.14159260) q[0];
    cx q[0],q[2];
    """

    res = transpile(prog, target="qasm2")

    # Normalize whitespace for comparison
    res_normalized = "".join(res.split())
    expected_normalized = "".join(expected.split())
    assert res_normalized == expected_normalized



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining exceptions for errors raised by qBraid.

"""
from typing import Any, Optional

from qbraid_core._import import LazyLoader

from qbraid.exceptions import QbraidError

registry = LazyLoader("registry", globals(), "qbraid.programs.registry")


class QasmError(QbraidError):
    """For errors raised while processing OpenQASM programs."""


class TransformError(QbraidError):
    """Base class for errors raised during qBraid transform processes."""


class PackageValueError(QbraidError):
    """Class for errors raised due to unsupported quantum frontend package"""

    def __init__(self, package: str):
        msg = (
            f"Quantum frontend module '{package}' is not supported.\n"
            f"Frontends supported by qBraid are: {registry.QPROGRAM_ALIASES}"
        )
        super().__init__(msg)


class ProgramTypeError(QbraidError):
    """Exception raised for errors encountered with unsupported quantum programs.

    Attributes:
        program (Optional[Any]): The program that caused the error, default is None.
        message (Optional[str]): Explanation of the error. If None, a default error message
                                 is generated based on the provided program.
    """

    def __init__(self, program: Optional[Any] = None, message: Optional[str] = None):
        self.program = program
        self.message = message
        super().__init__(self.generate_message())

    def generate_message(self) -> str:
        """Generate an error message based on the input."""
        if self.message is not None:
            return self.message
        if self.program is not None:
            return f"Quantum program of type '{type(self.program)}' is not supported."
        return "Unsupported quantum program type."


class ValidationError(ProgramTypeError):
    """Custom exception for validation errors in program types."""


class ProgramLoaderError(QbraidError):
    """Raised when an error occurs while loading a quantum program."""



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for generating dictionary of all qiskit gates for testing/benchmarking

"""

import string
from typing import Optional

import numpy as np
from qiskit.circuit.library import standard_gates as sg
from qiskit.circuit.library.standard_gates import *  # noqa: F403

qiskit_gates = {
    "H": sg.h.HGate,
    "X": sg.x.XGate,
    "Y": sg.y.YGate,
    "Z": sg.z.ZGate,
    "S": sg.s.SGate,
    "Sdg": sg.s.SdgGate,
    "T": sg.t.TGate,
    "Tdg": sg.t.TdgGate,
    "I": sg.i.IGate,
    "SX": sg.sx.SXGate,
    "SXdg": sg.sx.SXdgGate,
    "Phase": sg.p.PhaseGate,
    "RX": sg.rx.RXGate,
    "RY": sg.ry.RYGate,
    "RZ": sg.rz.RZGate,
    "U1": sg.U1Gate,
    "R": sg.r.RGate,
    "U2": sg.u2.U2Gate,
    "U": sg.u.UGate,
    "U3": sg.u3.U3Gate,
    "CH": sg.h.CHGate,
    "CX": sg.x.CXGate,
    "Swap": sg.swap.SwapGate,
    "iSwap": sg.iswap.iSwapGate,
    "CSX": sg.sx.CSXGate,
    "DCX": sg.dcx.DCXGate,
    "CY": sg.y.CYGate,
    "CZ": sg.z.CZGate,
    "CPhase": sg.p.CPhaseGate,
    "CRX": sg.rx.CRXGate,
    "RXX": sg.rxx.RXXGate,
    "CRY": sg.ry.CRYGate,
    "RYY": sg.ryy.RYYGate,
    "CRZ": sg.rz.CRZGate,
    "RZX": sg.rzx.RZXGate,
    "RZZ": sg.rzz.RZZGate,
    "CU1": sg.u1.CU1Gate,
    "RCCX": sg.x.RCCXGate,
    "CCX": sg.x.CCXGate,
}


def generate_params(varnames: list[str], seed: Optional[int] = None):
    """Returns a dictionary of random parameters for a given list of variable names"""
    if seed is not None:
        np.random.seed(seed)
    params = {
        ra: np.random.rand() * 2 * np.pi
        for ra in ["theta", "phi", "lam", "gamma"]
        if ra in varnames
    }
    if "num_ctrl_qubits" in varnames:
        params["num_ctrl_qubits"] = np.random.randint(1, 7)
    if "phase" in varnames:
        params["phase"] = np.random.rand() * 2 * np.pi
    return params


def get_qiskit_gates(seed: Optional[int] = None, exclude: list[str] = []):
    """Returns a dictionary of all qiskit gates with random parameters"""
    qiskit_gates = {attr: None for attr in dir(sg) if attr[0] in string.ascii_uppercase}
    gate_varnames = {}
    gate_equivs = {"MCXGrayCode": "MCXGate"}
    for gate in qiskit_gates:
        varnames = [v for v in getattr(sg, gate).__init__.__code__.co_varnames if v != "self"]
        if set(varnames).issubset({"args", "kwargs"}):
            if gate in gate_equivs and gate_equivs[gate] in gate_varnames:
                varnames = gate_varnames[gate_equivs[gate]]
            else:
                continue
        elif gate == "MCXGate":
            gate_varnames[gate] = varnames
        params = generate_params(varnames, seed=seed)
        qiskit_gates[gate] = getattr(sg, gate)(**params)
    return {k: v for k, v in qiskit_gates.items() if v is not None and k not in exclude}



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for generate random quantum circuits used for testing

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import numpy as np

from qbraid._logging import logger
from qbraid.exceptions import QbraidError
from qbraid.programs.exceptions import PackageValueError
from qbraid.programs.registry import QPROGRAM_ALIASES
from qbraid.transpiler.converter import transpile
from qbraid.transpiler.graph import ConversionGraph

if TYPE_CHECKING:
    import qbraid.programs


def random_circuit(
    package: str,
    num_qubits: Optional[int] = None,
    depth: Optional[int] = None,
    graph: Optional[ConversionGraph] = None,
    max_attempts: int = 1,
    **kwargs,
) -> qbraid.programs.QPROGRAM:
    """Generate random circuit of arbitrary size and form.

    Args:
        package (str): qBraid supported software package
        num_qubits (int, optional): Number of quantum wires.
            If not provided, set randomly in range [1,4].
        depth (int, optional): Layers of operations (i.e. critical path length)
            If not provided, set randomly in range [1,4].
        graph (ConversionGraph, optional): Conversion graph to use for transpilation
        max_attempts (int, optional): Maximum number of attempts to generate a random circuit
        **kwargs: Additional keyword arguments to pass to the random circuit generator

    Raises:
        PackageValueError: if ``package`` is not supported
        ValueError: when no conversion path exists for the specified package
        QbraidError: when random circuit generation fails for the specified package

    Returns:
        qbraid.programs.QPROGRAM: randomly generated quantum circuit/program
    """

    def validate_and_assign(value: Optional[int], name: str):
        if value is None:
            return np.random.randint(1, 4)
        if not isinstance(value, int) or value <= 0:
            raise ValueError(f"Invalid random circuit option. '{name}' must be a positive integer.")
        return value

    if package not in QPROGRAM_ALIASES:
        raise PackageValueError(f"Package '{package}' is not supported.")

    generator_funcs = {
        "ionq": "qbraid.interface.random.ionq_random.ionq_random",
        "qasm3": "qbraid.interface.random.qasm3_random.qasm3_random",
        "qiskit": "qbraid.interface.random.qiskit_random.qiskit_random",
        "cirq": "qbraid.interface.random.cirq_random.cirq_random",
    }
    graph = graph or ConversionGraph()
    valid_generators = [gen for gen in list(generator_funcs.keys()) if graph.has_path(gen, package)]

    if not valid_generators:
        raise ValueError(
            f"No registered generator that can create a random circuit for '{package}'."
        )

    num_qubits = validate_and_assign(num_qubits, "num_qubits")
    depth = validate_and_assign(depth, "depth")

    sorted_generators = graph.get_sorted_closest_sources(package, valid_generators)

    for src_pkg in sorted_generators:
        func_path = generator_funcs[src_pkg]
        try:
            module_name, func_name = func_path.rsplit(".", 1)
            module = __import__(module_name, fromlist=[func_name])
            rand_circuit_func = getattr(module, func_name)

            for attempt in range(max_attempts):
                try:
                    rand_circuit = rand_circuit_func(num_qubits, depth, **kwargs)
                    return transpile(rand_circuit, package)
                except Exception as attempt_err:  # pylint: disable=broad-exception-caught
                    if attempt < max_attempts - 1:
                        logger.info("Attempt %d failed: %s. Retrying...", attempt + 1, attempt_err)
                        continue
                    raise
        except Exception as err:  # pylint: disable=broad-exception-caught
            logger.info("Failed to generate circuit with %s: %s", src_pkg, err)
            continue

    raise QbraidError(f"Failed to generate random circuit for program type '{package}'.")


def random_unitary_matrix(dim: int) -> np.ndarray:
    """Create a random (complex) unitary matrix of order `dim`

    Args:
        dim: integer square matrix dimension

    Returns:
        random unitary matrix of shape dim x dim
    """
    # Create a random complex matrix of size dim x dim
    matrix = np.random.randn(dim, dim) + 1j * np.random.randn(dim, dim)
    # Use the QR decomposition to get a random unitary matrix
    unitary, _ = np.linalg.qr(matrix)
    return unitary



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining the ExperimentType enumeration.

"""
from __future__ import annotations

from enum import Enum


class ExperimentType(Enum):
    """
    Enumeration for quantum experiment types.

    Attributes:
        GATE_MODEL (str): Gate-based quantum computing (e.g., OpenQASM).
        AHS (str): Analog Hamiltonian simulation.
        ANNEALING (str): Quantum annealing for optimization problems.
        PHOTONIC (str): Photonic quantum computing using photons as qubits.
        OTHER (str): Placeholder for other or unspecified quantum computing models.
    """

    GATE_MODEL = "gate_model"
    AHS = "ahs"
    ANNEALING = "annealing"
    PHOTONIC = "photonic"
    OTHER = "other"



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for configuring provider credentials and authentication.

"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, overload

if TYPE_CHECKING:
    from qbraid.runtime.device import QuantumDevice


class QuantumProvider(ABC):
    """
    Abstract base class for managing interactions and authentication with various quantum
    service providers.

    This class defines the core interface that any quantum provider must implement to facilitate
    access to their quantum devices via qBraid Runtime. Subclasses should build upon this structure
    to incorporate any additional steps specific to their quantum platform, such as managing
    authentication, retrieving available devices, or handling other provider-specific functionality.

    Subclass implementations of the :py:meth:`QuantumProvider.get_device` and
    :py:meth:`QuantumProvider.get_devices` methods should dynamically construct a detailed
    :py:class:`TargetProfile` for each device. This profile is necessary to instantiate a
    :py:class:`QuantumDevice` and plays a central role in defining the runtime behavior and request
    structure necessary for submitting and executing quantum programs on that device through your
    service. To learn more, check out our documentation on
    `setting up a new provider <https://docs.qbraid.com/sdk/user-guide/runtime/new-provider>`_ and
    the `components of qBraid Runtime <https://docs.qbraid.com/sdk/user-guide/runtime/components>`_.

    """

    @overload
    def get_devices(self) -> list[QuantumDevice]: ...

    @overload
    def get_devices(self, **kwargs) -> list[QuantumDevice]: ...

    @abstractmethod
    def get_devices(self, **kwargs) -> list[QuantumDevice]:
        """
        Retrieves a list of available quantum devices, applying any specified filtering criteria.

        Args:
            **kwargs: Optional keyword arguments for filtering devices, if applicable.

        Returns:
            list[QuantumDevice]: A list of quantum devices that meet the filtering criteria,
                or all devices if no filters are applied.
        """

    @abstractmethod
    def get_device(self, device_id: str) -> QuantumDevice:
        """
        Retrieves the quantum device corresponding to the specified device ID.

        Args:
            device_id (str): The unique identifier of the quantum device.

        Returns:
            QuantumDevice: The quantum device matching the specified device ID.
        """

    def __eq__(self, other: Any) -> bool:
        """
        Compares two `QuantumProvider` instances for equality.

        By default, two `QuantumProvider` objects are considered equal if they are instances of the
        same class or subclass. Subclasses can override this behavior to include additional
        equality checks (e.g., comparing configuration or state).

        Args:
            other (Any): The object to compare with the current instance.

        Returns:
            bool: True if both objects are instances of the same class or subclass; False otherwise.
        """
        if not isinstance(other, type(self)):
            return False
        return True



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module providing utilities for interfacing with supported quantum programs.

.. currentmodule:: qbraid.interface

Functions
-----------

.. autosummary::
   :toctree: ../stubs/

   circuits_allclose
   assert_allclose_up_to_global_phase

"""
from .circuit_equality import assert_allclose_up_to_global_phase, circuits_allclose
from .random import random_circuit, random_unitary_matrix

__all__ = [
    "assert_allclose_up_to_global_phase",
    "circuits_allclose",
    "random_circuit",
    "random_unitary_matrix",
]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining QbraidJob class

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type, Union

from qbraid_core.services.quantum import QuantumClient

from qbraid._logging import logger
from qbraid.programs import ExperimentType
from qbraid.runtime.enums import JobStatus
from qbraid.runtime.exceptions import JobStateError, QbraidRuntimeError
from qbraid.runtime.job import QuantumJob
from qbraid.runtime.result import Result, ResultDataType
from qbraid.runtime.result_data import AhsResultData, AnnealingResultData, GateModelResultData
from qbraid.runtime.schemas import RuntimeJobModel

from .result import (
    Equal1SimulatorResultData,
    NECVectorAnnealerResultData,
    QbraidQirSimulatorResultData,
    QuEraQasmSimulatorResultData,
)

if TYPE_CHECKING:
    import qbraid_core.services.quantum

    import qbraid.runtime


class QbraidJob(QuantumJob):
    """Class representing a qBraid job."""

    def __init__(
        self,
        job_id: str,
        device: Optional[qbraid.runtime.QbraidDevice] = None,
        client: Optional[qbraid_core.services.quantum.QuantumClient] = None,
        **kwargs,
    ):
        super().__init__(job_id, device, **kwargs)
        self._client = client

    @property
    def client(self) -> qbraid_core.services.quantum.QuantumClient:
        """
        Lazily initializes and returns the client object associated with the job.
        If the job has an associated device with a client, that client is used.
        Otherwise, a new instance of QuantumClient is created and used.

        Returns:
            QuantumClient: The client object associated with the job.
        """
        if self._client is None:
            self._client = self._device.client if self._device else QuantumClient()
        return self._client

    def queue_position(self) -> Optional[int]:
        """Return the position of the job in the queue."""
        job_data = self.metadata()
        return job_data.get("queue_position", job_data.get("queuePosition"))

    def status(self) -> JobStatus:
        """Return the status of the job / task , among the values of ``JobStatus``."""
        terminal_states = JobStatus.terminal_states()
        if self._cache_metadata.get("status") not in terminal_states:
            client_data = self.client.get_job(self.id)
            job_model = RuntimeJobModel.from_dict(client_data)
            job_data = job_model.model_dump(exclude={"metadata", "cost"})
            status = JobStatus(job_data.pop("status"))
            if job_model.status_text is not None:
                status.set_status_message(job_model.status_text)
            self._cache_metadata.update({**job_data, "status": status})
        return self._cache_metadata["status"]

    def cancel(self) -> None:
        """Attempt to cancel the job."""
        if self.is_terminal_state():
            raise JobStateError("Cannot cancel job in a terminal state.")

        self.client.cancel_job(self.id)
        logger.info("Cancel job request validated.")

        try:
            logger.info("Waiting for job to cancel...")
            self.wait_for_final_state(timeout=3, poll_interval=1)
        except TimeoutError:
            pass

        status = self.status()
        if status not in {JobStatus.CANCELLED, JobStatus.CANCELLING}:
            raise QbraidRuntimeError(f"Failed to cancel job. Current status: {status.name}")

        logger.info("Success. Current status: %s", status.name)

    @staticmethod
    def get_result_data_cls(
        device_id: Optional[str] = None, experiment_type: Optional[ExperimentType] = None
    ) -> Union[Type[GateModelResultData], Type[AnnealingResultData], Type[AhsResultData]]:
        """Determine the appropriate ResultData class based on device_id and experiment_type."""
        device_to_result_data = {
            "qbraid_qir_simulator": QbraidQirSimulatorResultData,
            "quera_qasm_simulator": QuEraQasmSimulatorResultData,
            "nec_vector_annealer": NECVectorAnnealerResultData,
            "equal1_simulator": Equal1SimulatorResultData,
        }

        result_data_cls = device_to_result_data.get(device_id)

        if not result_data_cls:
            experiment_type_to_result_data = {
                ExperimentType.GATE_MODEL: GateModelResultData,
                ExperimentType.ANNEALING: AnnealingResultData,
                ExperimentType.AHS: AhsResultData,
            }
            result_data_cls = experiment_type_to_result_data.get(experiment_type)

        if not result_data_cls:
            raise ValueError(
                f"Unsupported device_id '{device_id}' or experiment_type '{experiment_type.name}'"
            )

        return result_data_cls

    def result(self, timeout: Optional[int] = None) -> Result[ResultDataType]:
        """Return the results of the job."""
        self.wait_for_final_state(timeout=timeout)
        job_data = self.client.get_job(self.id)
        success = job_data.get("status") == JobStatus.COMPLETED.name
        job_result = (
            self.client.get_job_results(self.id, wait_time=1, backoff_factor=1.4) if success else {}
        )
        job_result.update(job_data)
        model = RuntimeJobModel.from_dict(job_result)
        result_data_cls = self.get_result_data_cls(model.device_id, model.experiment_type)
        data = result_data_cls.from_object(model.metadata)
        exclude = (
            {"solutions", "num_solutions"}
            if model.experiment_type == ExperimentType.ANNEALING
            else {"measurement_counts", "measurements"}
        )
        metadata_dump = model.metadata.model_dump(by_alias=True, exclude=exclude)
        model_dump = model.model_dump(
            by_alias=True, exclude={"job_id", "device_id", "metadata", "queue_position"}
        )
        experiment_type: ExperimentType = model_dump["experimentType"]
        status_text = (
            model.status_text or model.status.status_message or model.status.default_message
        )
        model_dump.update(
            {
                "status": model.status.name,
                "statusText": status_text,
                "experimentType": experiment_type,
            }
        )
        model_dump["metadata"] = metadata_dump
        return Result[ResultDataType](
            device_id=model.device_id, job_id=model.job_id, success=success, data=data, **model_dump
        )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing Cirq circuits used for testing

"""
import random
from typing import Optional

import numpy as np
from cirq import (
    CNOT,
    ISWAP,
    SWAP,
    Circuit,
    CZPowGate,
    H,
    LineQubit,
    S,
    T,
    X,
    XPowGate,
    Y,
    Z,
    ZPowGate,
    measure,
    ops,
    rx,
    ry,
    rz,
)


def cirq_bell() -> Circuit:
    """Returns Cirq bell circuit"""
    q0, q1 = LineQubit.range(2)
    circuit = Circuit(ops.H(q0), ops.CNOT(q0, q1))
    return circuit


def cirq_shared15():
    """Returns cirq `Circuit` for qBraid `TestSharedGates`."""

    circuit = Circuit()
    q0, q1, q2, q3 = LineQubit.range(4)

    cirq_shared_gates = [
        H(q0),
        H(q1),
        H(q2),
        H(q3),
        X(q0),
        X(q1),
        Y(q2),
        Z(q3),
        S(q0),
        ZPowGate(exponent=-0.5)(q1),
        T(q2),
        ZPowGate(exponent=-0.25)(q3),
        rx(rads=np.pi / 4)(q0),
        ry(rads=np.pi / 2)(q1),
        rz(rads=3 * np.pi / 4)(q2),
        ZPowGate(exponent=1 / 8)(q3),
        XPowGate(exponent=0.5)(q0),
        XPowGate(exponent=-0.5)(q1),
        ISWAP(q2, q3),
        SWAP(q0, q2),
        SWAP(q1, q3),
        CNOT(q0, q1),
        CZPowGate(exponent=0.25)(q2, q3),
    ]

    for gate in cirq_shared_gates:
        circuit.append(gate)

    return circuit


def sparse_circuit(num_qubits: Optional[int] = None) -> Circuit:
    """
    Generates a quantum circuit designed to benchmark the performance of a sparse simulator.

    This circuit is structured to maintain a level of sparsity in the system's state vector, making
    it a good candidate for testing sparse quantum simulators. Sparse simulators excel in
    simulating circuits where the state vector remains sparse, i.e., most of its elements are zero
    or can be efficiently represented.

    Args:
        num_qubits (optional, int): The number of qubits to use in the circuit. If not provided,
                                    a random number of qubits between 10 and 20 will be used.

    Returns:
        cirq.Circuit: The constructed circuit for benchmarking
    """
    num_qubits = num_qubits or random.randint(10, 20)
    # Create a circuit
    circuit = Circuit()

    # Create qubits
    qubits = LineQubit.range(num_qubits)

    # Apply Hadamard gates to the first half of the qubits
    for qubit in qubits[: num_qubits // 2]:
        circuit.append(H(qubit))

    # Apply a CNOT ladder
    for i in range(num_qubits - 1):
        circuit.append(CNOT(qubits[i], qubits[i + 1]))

    # Apply Z gates to randomly selected qubits
    for qubit in random.sample(qubits, k=num_qubits // 2):
        circuit.append(Z(qubit))

    # Measurement (optional)
    circuit.append(measure(*qubits, key="result"))

    return circuit



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for BraketProvider (remote)

"""
from decimal import Decimal

import pytest
from braket.aws.aws_device import AwsDevice
from braket.circuits import Circuit
from braket.tracking.tracker import Tracker

from qbraid.runtime.aws.provider import BraketProvider
from qbraid.runtime.aws.tracker import get_quantum_task_cost


@pytest.fixture
def braket_no_meas(braket_circuit: Circuit) -> Circuit:
    """Return a circuit without measurements."""
    circuit = Circuit()

    for instr in braket_circuit.instructions:
        if str(instr.operator) != "Measure":
            circuit.add_instruction(instr)

    return circuit


@pytest.fixture
def braket_most_busy():
    """Return the most busy device for testing purposes."""
    provider = BraketProvider()
    qbraid_devices = provider.get_devices(
        types=["QPU"], statuses=["ONLINE"], provider_names=["Rigetti", "IonQ", "Oxford", "IQM"]
    )

    qbraid_device = None
    max_queued = 0
    for device in qbraid_devices:
        jobs_queued = device.queue_depth()
        if jobs_queued is not None and jobs_queued > max_queued:
            max_queued = jobs_queued
            qbraid_device = device
    return qbraid_device


@pytest.mark.remote
def test_get_quantum_task_cost_simulator(braket_circuit):
    """Test getting cost of quantum task run on an AWS simulator."""
    provider = BraketProvider()
    device = provider.get_device("arn:aws:braket:::device/quantum-simulator/amazon/sv1")

    with Tracker() as tracker:
        task = device.run(braket_circuit, shots=2)
        task.result()

    expected = tracker.simulator_tasks_cost()
    calculated = get_quantum_task_cost(task.id, provider._get_aws_session())
    assert expected == calculated


@pytest.mark.remote
def test_get_quantum_task_cost_cancelled(braket_most_busy, braket_no_meas):
    """Test getting cost of quantum task that was cancelled."""
    if braket_most_busy is None:
        pytest.skip("No AWS QPU devices available")

    provider = BraketProvider()

    # AwsSession region must match device region
    region_name = AwsDevice.get_device_region(braket_most_busy.id)
    aws_session = provider._get_aws_session(region_name)

    qbraid_job = braket_most_busy.run(braket_no_meas, shots=10)
    qbraid_job.cancel()

    task_arn = qbraid_job.id

    try:
        qbraid_job.wait_for_final_state(timeout=30)
        final_state_reached = True
    except TimeoutError:
        final_state_reached = False

    # Based on whether final state was reached or not, proceed to verify expected outcomes
    if final_state_reached:
        # Verify cost is as expected when job reaches a final state
        cost = get_quantum_task_cost(task_arn, aws_session)
        assert cost == Decimal(0), f"Expected cost to be 0 when job is in a final state, got {cost}"
    else:
        # Verify the appropriate error is raised when job has not reached a final state
        with pytest.raises(ValueError) as excinfo:
            get_quantum_task_cost(task_arn, aws_session)

        expected_msg_partial = f"Task {task_arn} is not COMPLETED."
        assert expected_msg_partial in str(
            excinfo.value
        ), "Unexpected error message for non-final job state"


@pytest.mark.remote
def test_get_quantum_task_cost_region_mismatch(braket_most_busy, braket_circuit):
    """Test getting cost of quantum task raises value error on region mismatch."""
    if braket_most_busy is None:
        pytest.skip("No AWS QPU devices available")

    braket_device = braket_most_busy._device
    task = braket_device.run(braket_circuit, shots=10)
    task.cancel()

    task_arn = task.id
    task_region = task_arn.split(":")[3]
    other_region = "eu-west-2" if task_region == "us-east-1" else "us-east-1"

    provider = BraketProvider()
    aws_session = provider._get_aws_session(other_region)

    with pytest.raises(ValueError) as excinfo:
        get_quantum_task_cost(task_arn, aws_session)

    assert (
        str(excinfo.value)
        == f"AwsSession region {other_region} does not match task region {task_region}"
    )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qbraid.programs.annealing module.

"""
import json
from unittest.mock import Mock

import pytest

try:
    from pyqubo import Spin

    from qbraid.programs import ExperimentType, ProgramTypeError, unregister_program_type
    from qbraid.programs.annealing import (
        AnnealingProgram,
        Problem,
        ProblemEncoder,
        ProblemType,
        QuboProblem,
    )
    from qbraid.programs.annealing.cpp_pyqubo import PyQuboModel
    from qbraid.programs.annealing.qubo import QuboProgram
    from qbraid.runtime.native.provider import _serialize_program

    pyqubo_not_installed = False
except ImportError:
    pyqubo_not_installed = True


pytestmark = pytest.mark.skipif(pyqubo_not_installed, reason="pyqubo not installed")


@pytest.fixture
def pyqubo_model():
    """Creates and returns a compiled pyqubo model using Spin variables."""
    s1, s2, s3, s4 = Spin("s1"), Spin("s2"), Spin("s3"), Spin("s4")
    H = (4 * s1 + 2 * s2 + 7 * s3 + s4) ** 2
    model = H.compile()
    return model


@pytest.fixture
def mock_annealing_program():
    """Creates and returns a dummy subclass of AnnealingProgram for testing."""

    class MockAnnealingProgram(AnnealingProgram):
        """A dummy subclass of AnnealingProgram for testing."""

        # pylint: disable-next=super-init-not-called
        def __init__(self, problem: Problem):
            self._problem = problem

        def to_problem(self) -> Problem:
            return self._problem

    return MockAnnealingProgram


def test_problem_initialization():
    """Tests the initialization of a Problem instance with specified
    linear and quadratic coefficients."""
    problem = Problem(
        problem_type=ProblemType.QUBO,
        linear={"x1": 1.0, "x2": -1.5},
        quadratic={("x1", "x2"): 0.5},
    )

    assert problem.problem_type == ProblemType.QUBO
    assert problem.linear == {"x1": 1.0, "x2": -1.5}
    assert problem.quadratic == {("x1", "x2"): 0.5}
    assert problem.num_variables() == 2


def test_problem_equality():
    """Tests the equality comparison between two Problem instances with
    identical properties."""
    problem1 = Problem(
        problem_type=ProblemType.QUBO,
        linear={"x1": 1.0},
        quadratic={("x1", "x2"): 0.5},
    )
    problem2 = Problem(
        problem_type=ProblemType.QUBO,
        linear={"x1": 1.0},
        quadratic={("x2", "x1"): 0.5},
    )

    assert problem1 == problem2


def test_qubo_problem_initialization():
    """Tests the initialization of a QuboProblem instance with given
    quadratic coefficients and offset."""
    coefficients = {("x1", "x2"): 0.5, ("x2", "x3"): -1.0}
    qubo_problem = QuboProblem(coefficients)

    assert qubo_problem.problem_type == ProblemType.QUBO
    assert qubo_problem.quadratic == coefficients
    assert qubo_problem.num_variables() == 3


def test_pyqubo_model_initialization(pyqubo_model):
    """Tests the initialization of a PyQuboModel instance using a pyqubo model fixture."""
    pyqubo_model_instance = PyQuboModel(pyqubo_model)

    assert pyqubo_model_instance.program == pyqubo_model


def get_program_classes():
    """Dynamically import and return program classes for testing."""
    try:
        return [PyQuboModel, QuboProgram]
    except NameError:
        return []


@pytest.mark.parametrize("program_class", get_program_classes())
def test_invalid_program_initialization(program_class):
    """Tests that initializing a program with an invalid input raises a ProgramTypeError."""
    try:
        with pytest.raises(ProgramTypeError):
            program_class(("invalid", "program"))
    finally:
        unregister_program_type("tuple")


def test_pyqubo_model_to_problem(pyqubo_model):
    """Tests the conversion of a PyQuboModel instance to a QuboProblem."""
    pyqubo_model_instance = PyQuboModel(pyqubo_model)

    problem = pyqubo_model_instance.to_problem()
    assert isinstance(problem, QuboProblem)
    assert problem.num_variables() == 4
    assert pyqubo_model_instance.num_qubits == 4


def test_problem_encoder(mock_annealing_program):
    """Tests the ProblemEncoder class for encoding Problem instances to JSON."""
    problem = Problem(
        problem_type=ProblemType.QUBO,
        linear={"x1": 1.0, "x2": -1.5},
        quadratic={("x1", "x2"): 0.5},
    )

    problem_json = json.dumps(problem, cls=ProblemEncoder)
    expected_problem_json = json.dumps(
        {
            "problem_type": "qubo",
            "linear": {"x1": 1.0, "x2": -1.5},
            "quadratic": {'["x1", "x2"]': 0.5},
        }
    )
    assert problem_json == expected_problem_json

    mock_program = mock_annealing_program(problem)

    program_json = json.dumps(mock_program, cls=ProblemEncoder)
    assert program_json == expected_problem_json == mock_program.to_json()

    problem_type_json = json.dumps(ProblemType.QUBO, cls=ProblemEncoder)
    assert problem_type_json == '"qubo"'


def test_problem_eq_different_type():
    """Test __eq__ returns False when compared with a non-Problem instance."""
    problem = Problem(problem_type=ProblemType.QUBO)
    assert problem != "Not a Problem instance"


def test_problem_eq_different_problem_type():
    """Test __eq__ returns False when problem_types are different."""
    problem1 = Problem(problem_type=ProblemType.QUBO)
    problem2 = Problem(problem_type=ProblemType.ISING)
    assert problem1 != problem2


def test_problem_eq_different_linear():
    """Test __eq__ returns False when linear terms are different."""
    problem1 = Problem(problem_type=ProblemType.QUBO, linear={"x1": 1.0})
    problem2 = Problem(problem_type=ProblemType.QUBO, linear={"x1": 2.0})
    assert problem1 != problem2


def test_problem_eq_different_quadratic_length():
    """Test __eq__ returns False when quadratic term lengths are different."""
    problem1 = Problem(problem_type=ProblemType.QUBO, quadratic={("x1", "x2"): 0.5})
    problem2 = Problem(problem_type=ProblemType.QUBO, quadratic={})
    assert problem1 != problem2


def test_problem_eq_quadratic_key_not_in_other():
    """Test __eq__ returns False when a quadratic key is missing in the other problem."""
    problem1 = Problem(problem_type=ProblemType.QUBO, quadratic={("x1", "x2"): 0.5})
    problem2 = Problem(problem_type=ProblemType.QUBO, quadratic={("x2", "x3"): 0.5})
    assert problem1 != problem2


def test_problem_eq_quadratic_value_differs():
    """Test __eq__ returns False when quadratic values differ for the same key."""
    problem1 = Problem(problem_type=ProblemType.QUBO, quadratic={("x1", "x2"): 0.5})
    problem2 = Problem(problem_type=ProblemType.QUBO, quadratic={("x1", "x2"): 1.0})
    assert problem1 != problem2


def test_problem_eq_quadratic_keys_swapped_same_value():
    """Test __eq__ returns True when quadratic keys are swapped but values are the same."""
    problem1 = Problem(problem_type=ProblemType.QUBO, quadratic={("x1", "x2"): 0.5})
    problem2 = Problem(problem_type=ProblemType.QUBO, quadratic={("x2", "x1"): 0.5})
    assert problem1 == problem2


def test_problem_eq_quadratic_keys_swapped_value_differs():
    """Test __eq__ returns False when quadratic keys are swapped and values differ."""
    problem1 = Problem(problem_type=ProblemType.QUBO, quadratic={("x1", "x2"): 0.5})
    problem2 = Problem(problem_type=ProblemType.QUBO, quadratic={("x2", "x1"): 1.0})
    assert problem1 != problem2


def test_problem_eq_all_attributes_same():
    """Test __eq__ returns True when all attributes are identical."""
    problem1 = Problem(
        problem_type=ProblemType.QUBO,
        linear={"x1": 1.0, "x2": -1.5},
        quadratic={("x1", "x2"): 0.5, ("x3", "x4"): -0.3},
    )
    problem2 = Problem(
        problem_type=ProblemType.QUBO,
        linear={"x1": 1.0, "x2": -1.5},
        quadratic={("x4", "x3"): -0.3, ("x2", "x1"): 0.5},
    )
    assert problem1 == problem2


def test_annealing_program_eq(mock_annealing_program):
    """Test __eq__ returns True when compared with an equivalent instance."""
    problem = Problem(problem_type=ProblemType.QUBO)
    annealing_program = mock_annealing_program(problem)
    annealing_program_2 = mock_annealing_program(problem)
    assert annealing_program == annealing_program_2


def test_annealing_program_eq_different_type(mock_annealing_program):
    """Test __eq__ returns False when compared with a non-AnnealingProgram instance."""
    problem = Problem(problem_type=ProblemType.QUBO)
    annealing_program = mock_annealing_program(problem)
    assert annealing_program != "Not an AnnealingProgram instance"


def test_runtime_serliaze_qubo(pyqubo_model):
    """Test that the _pyqubo_to_json function returns the expected JSON string."""
    pyqubo_json = _serialize_program(pyqubo_model)
    pyqubo_dict = {"problem": json.loads(pyqubo_json["problem"])}

    expected_dict = {
        "problem": {
            "problem_type": "qubo",
            "quadratic": {
                '["s1", "s1"]': -160.0,
                '["s4", "s2"]': 16.0,
                '["s3", "s1"]': 224.0,
                '["s2", "s2"]': -96.0,
                '["s4", "s1"]': 32.0,
                '["s1", "s2"]': 64.0,
                '["s3", "s2"]': 112.0,
                '["s3", "s3"]': -196.0,
                '["s4", "s4"]': -52.0,
                '["s4", "s3"]': 56.0,
            },
        }
    }

    assert pyqubo_dict == expected_dict


def test_problem_encoder_super_default():
    """Test that the ProblemEncoder falls back to the default encoder."""
    data = {"data": "test_data"}
    encoded_obj = json.dumps(data, cls=ProblemEncoder)
    expected_output = '{"data": "test_data"}'
    assert encoded_obj == expected_output


def test_get_pyqubo_experiment_type(pyqubo_model):
    """Test that the PyQuboModel correctly identifies the experiment type as ANNEALING."""
    program = PyQuboModel(pyqubo_model)
    assert program.experiment_type == ExperimentType.ANNEALING


def test_pyqubo_model_transform(pyqubo_model):
    """Test that the PyQuboModel transform method does not modify the program."""
    program = PyQuboModel(pyqubo_model)
    program.transform(device=Mock())
    assert program.program == pyqubo_model



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for interfacing quantum programs

"""
from unittest.mock import patch

import pytest

from qbraid._version import __version__
from qbraid.interface import circuits_allclose, random_circuit
from qbraid.interface.random.cirq_random import cirq_random
from qbraid.interface.random.qasm3_random import qasm3_random
from qbraid.interface.random.qiskit_random import qiskit_random
from qbraid.programs.exceptions import QbraidError
from qbraid.transpiler import ConversionGraph, transpile


@pytest.mark.parametrize("num_qubits, depth, max_operands, seed", [(1, 1, 1, 42)])
def test_qasm3_random(num_qubits, depth, max_operands, seed):
    """Test that qasm3_random generates the correct QASM code."""
    expected_output = f"""
// Generated by qBraid v{__version__}
OPENQASM 3.0;
include "stdgates.inc";
/*
    seed = {seed}
    num_qubits = {num_qubits}
    depth = {depth}
    max_operands = {max_operands}
*/
qubit[1] q;
bit[1] c;
x q[0];
c[0] = measure q[0];
"""
    output = random_circuit(
        "qasm3",
        num_qubits=num_qubits,
        depth=depth,
        max_operands=max_operands,
        seed=seed,
        measure=True,
    )
    assert output == expected_output


def test_badqasm3_random():
    """Test that qasm3_random raises a QbraidError when it fails."""
    with pytest.raises(QbraidError):
        random_circuit("qasm3", seed="12")


@pytest.mark.parametrize("param", ["num_qubits", "depth"])
def test_qasm3_zero_value_raises(param):
    """Test that qasm3_random raises a ValueError when a circuit parameter is <=0."""
    params = {param: 0}
    expected_err = f"Invalid random circuit option. '{param}' must be a positive integer."
    with pytest.raises(ValueError, match=expected_err):
        random_circuit("qasm3", **params)


@pytest.mark.parametrize("param", ["max_operands"])
def test_qasm3_zero_value_raises_for_max_operands(param):
    """Test that qasm3_random raises a ValueError when a circuit parameter is <=0."""
    expected_err = f"Invalid random circuit option. '{param}' must be a positive integer."
    with pytest.raises(ValueError, match=expected_err):
        qasm3_random(max_operands=0)


@pytest.mark.parametrize("package", ["qiskit", "cirq"])
def test_random_circuit_raises_for_bad_params(package: str, available_targets):
    """Test that cirq_random raises a QbraidError for invalid parameters."""
    if package not in available_targets:
        pytest.skip(f"{package} not installed")

    err_msg = f"Failed to generate random circuit for program type '{package}'."
    with pytest.raises(QbraidError, match=err_msg):
        random_circuit(package, made_up_param=42)


def test_circuits_allclose(available_targets):
    """Test circuit allclose function."""
    pytket_available = "pytket" in available_targets
    braket_available = "braket" in available_targets
    qiskit_available = "qiskit" in available_targets

    if not (pytket_available and (braket_available or qiskit_available)):
        pytest.skip("Required quantum package(s) are not available")

    circuit0 = random_circuit("pytket", num_qubits=2, depth=2)

    if braket_available:
        circuit1 = transpile(circuit0, "braket")
        assert circuits_allclose(circuit1, circuit0, index_contig=True, allow_rev_qubits=True)

    if qiskit_available:
        circuit2 = random_circuit("qiskit", num_qubits=3, depth=2)
        assert not circuits_allclose(circuit2, circuit0, index_contig=True, allow_rev_qubits=True)


def test_bad_random_circuit():
    """Test that random_circuit raises a PackageValueError when given a bad package."""
    with pytest.raises(QbraidError):
        random_circuit("bad_package")


def test_raise_value_error_no_valid_generators():
    """Test raising ValueError when no valid generators are available"""
    with pytest.raises(
        ValueError, match="No registered generator that can create a random circuit for 'qasm2'"
    ):
        random_circuit("qasm2", graph=ConversionGraph(nodes=["qasm2", "qasm3"]))


def test_cirq_random_raises_for_bad_param():
    """Test that cirq_random raises a ValueError when a circuit parameter is <=0."""
    with pytest.raises(QbraidError, match="Failed to create Cirq random circuit"):
        cirq_random(1, 2, op_density=-1)


def test_qiskit_random_raises_for_bad_param():
    """Test that qiskit_random raises a QbraidError when a circuit parameter is <=0."""
    with pytest.raises(QbraidError, match="Failed to create Qiskit random circuit"):
        qiskit_random(1, 2, max_operands=-1)


@pytest.mark.parametrize("max_attempts", [1, 2, 3])
def test_random_circuit_max_attempts(max_attempts):
    """Test that random_circuit raises a QbraidError when the max number of attempts is reached."""
    with patch("qbraid.interface.random.random.transpile") as mock_transpile:
        mock_transpile.side_effect = Exception("Mocked transpile exception")

        with pytest.raises(QbraidError):
            random_circuit("qiskit", num_qubits=2, depth=2, max_attempts=max_attempts)

        assert mock_transpile.call_count == max_attempts * 3



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for qbraid.programs.pytket.PytketCircuit

"""
from unittest.mock import patch

import numpy as np
import pytest

try:
    from pytket.circuit import Circuit, OpType
    from pytket.unit_id import Qubit

    from qbraid.programs.exceptions import ProgramTypeError, TransformError
    from qbraid.programs.gate_model.pytket import IONQ_GATES, PytketCircuit

    pytket_not_installed = False
except ImportError:
    pytket_not_installed = True


pytestmark = pytest.mark.skipif(pytket_not_installed, reason="pytket not installed")


def test_program_attributes():
    """Test attributes of pytket circuit"""
    circuit = Circuit(2, 1)
    circuit.H(0)
    circuit.CX(0, 1)
    qprogram = PytketCircuit(circuit)
    assert qprogram.qubits == [Qubit(0), Qubit(1)]
    assert qprogram.num_qubits == 2
    assert qprogram.num_clbits == 1
    assert qprogram.depth == 2


@pytest.fixture
def circuits():
    """Fixture to generate circuits only when called."""
    return {
        "simple": (Circuit(2).CX(0, 1), Circuit(2).CX(1, 0)),
        "ccx": (Circuit(3).CCX(0, 1, 2), Circuit(3).CCX(2, 1, 0)),
    }


@pytest.mark.parametrize("circuit_key", ["simple", "ccx"])
def test_reverse_qubit_order(circuits, circuit_key):
    """Test reversing qubit ordering of pytket circuit using a fixture."""
    input_circuit, expected_circuit = circuits[circuit_key]
    qprogram = PytketCircuit(input_circuit)
    qprogram.reverse_qubit_order()
    result_circuit = qprogram.program
    assert result_circuit.get_commands() == expected_circuit.get_commands()


def test_remove_idle_qubits_pytket():
    """Test wires with no operations from pytket circuit"""
    circuit = Circuit(3)
    circuit.H(0)
    circuit.CX(0, 1)
    qprogram = PytketCircuit(circuit)
    qprogram.remove_idle_qubits()
    contig_circuit = qprogram.program
    assert contig_circuit.n_qubits == 2


def test_remove_measurements():
    """Test removing measurements from pytket circuit"""
    circuit = Circuit(2, 1)
    circuit.H(0)
    circuit.CX(0, 1)
    circuit.Measure(0, 0)
    qprogram = PytketCircuit(circuit)
    new_circuit = qprogram.remove_measurements(qprogram.program)

    for command in new_circuit.get_commands():
        assert command.op.type != OpType.Measure


@pytest.mark.parametrize("flat", [True, False])
@pytest.mark.parametrize("list_type", [True, False])
def test_gate_to_matrix_pytket(flat, list_type):
    """Test converting pytket gates to matrix"""
    c = Circuit(10, 2, name="example")
    c.CU1(np.pi / 2, 2, 3)

    c_unitary = PytketCircuit.gate_to_matrix(
        gates=c.get_commands()[0] if list_type else c.get_commands(), flat=flat
    )
    if flat:
        assert c_unitary.shape[0] == 2**2
    else:
        assert c_unitary.shape[0] == 2**4


def test_assertion_error_in_rebase():
    """Test assertion error in rebase method"""
    circuit = Circuit(1).H(0)
    gates = IONQ_GATES
    max_qubits = 5

    with patch(
        "qbraid.programs.gate_model.pytket.CompilationUnit.check_all_predicates", return_value=False
    ):
        with pytest.raises(TransformError) as excinfo:
            PytketCircuit.rebase(circuit, gates, max_qubits)

    assert "Rebased circuit failed to satisfy compilation predicates" in str(excinfo.value)


def test_raise_program_type_error():
    """Test raising ProgramTypeError"""
    with pytest.raises(ProgramTypeError):
        PytketCircuit("OPENQASM 2.0;qreg q[2];h q[0];cx q[0],q[1];")



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qasm2/qasm3 to IonQDictType transpilation

"""
import importlib.util
import sys
from unittest.mock import Mock, patch

import openqasm3.ast
import pytest
from openqasm3.parser import parse

from qbraid.programs.gate_model.ionq import GateSet, InputFormat
from qbraid.programs.gate_model.qasm3 import OpenQasm3Program
from qbraid.programs.typer import IonQDictType, Qasm3StringType
from qbraid.transpiler.conversions.openqasm3.openqasm3_to_ionq import (
    _parse_gates,
    extract_params,
    openqasm3_to_ionq,
)
from qbraid.transpiler.conversions.qasm2.qasm2_to_ionq import qasm2_to_ionq
from qbraid.transpiler.conversions.qasm3.qasm3_to_ionq import qasm3_to_ionq
from qbraid.transpiler.exceptions import ProgramConversionError


def test_ionq_device_extract_gate_data():
    """Test extracting gate data from a OpenQASM 2 program."""
    qasm = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[3];
    x q[0];
    not q[1];
    y q[0];
    z q[0], q[1];
    rx(pi / 4) q[0];
    ry(pi / 2) q[0];
    rz(3 * pi / 4) q[0];
    h q[0];
    cx q[0], q[1];
    CX q[1], q[2];
    cnot q[2], q[0];
    ccx q[0], q[1], q[2];
    toffoli q[2], q[1], q[0];
    s q[0];
    sdg q[0];
    si q[0];
    id q[0];
    t q[0];
    tdg q[0];
    ti q[1];
    sx q[0];
    v q[1];
    sxdg q[0];
    vi q[1];
    swap q[0], q[1];
    """

    gate_data = [
        {"gate": "x", "target": 0},
        {"gate": "not", "target": 1},
        {"gate": "y", "target": 0},
        {"gate": "z", "target": 0},
        {"gate": "z", "target": 1},
        {"gate": "rx", "target": 0, "rotation": 0.7853981633974483},
        {"gate": "ry", "target": 0, "rotation": 1.5707963267948966},
        {"gate": "rz", "target": 0, "rotation": 2.356194490192345},
        {"gate": "h", "target": 0},
        {"gate": "cnot", "control": 0, "target": 1},
        {"gate": "cnot", "control": 1, "target": 2},
        {"gate": "cnot", "control": 2, "target": 0},
        {"gate": "cnot", "controls": [0, 1], "target": 2},
        {"gate": "cnot", "controls": [2, 1], "target": 0},
        {"gate": "s", "target": 0},
        {"gate": "si", "target": 0},
        {"gate": "si", "target": 0},
        {"gate": "t", "target": 0},
        {"gate": "ti", "target": 0},
        {"gate": "ti", "target": 1},
        {"gate": "v", "target": 0},
        {"gate": "v", "target": 1},
        {"gate": "vi", "target": 0},
        {"gate": "vi", "target": 1},
        {"gate": "swap", "targets": [0, 1]},
    ]
    expected = {
        "qubits": 3,
        "circuit": gate_data,
        "gateset": GateSet.QIS.value,
        "format": InputFormat.CIRCUIT.value,
    }

    actual = qasm2_to_ionq(qasm)

    assert actual == expected


def test_qasm2_to_ionq_measurements_raises():
    """Test that qasm2_to_ionq emits warning when the circuit contains measurements."""
    qasm = """
    OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[1];
    creg c[1];
    x q[0];
    measure q[0] -> c[0];
    """
    with pytest.warns(
        UserWarning,
        match=(
            "Circuit contains measurement gates, which will be ignored "
            "during conversion to the IonQDictType"
        ),
    ):
        qasm2_to_ionq(qasm)


@pytest.fixture
def deutsch_jozsa_qasm3() -> Qasm3StringType:
    """Return a QASM 3.0 string for the DJ algorithm."""
    return """
    OPENQASM 3.0;
    include "stdgates.inc";

    gate hgate q { h q; }
    gate xgate q { x q; }

    const int[32] N = 4;
    qubit[4] q;
    qubit ancilla;

    def deutsch_jozsa(qubit[N] q_func, qubit[1] ancilla_q) {
    xgate ancilla_q;
    for int i in [0:N-1] { hgate q_func[i]; }
    hgate ancilla_q;
    for int i in [0:N-1] { cx q_func[i], ancilla_q; }
    for int i in [0:N-1] { hgate q_func[i]; }
    }

    deutsch_jozsa(q, ancilla);
    """


@pytest.fixture
def deutch_jozsa_qasm3_unrolled() -> Qasm3StringType:
    """Return an unrolled QASM 3.0 string for the DJ algorithm."""
    return """
    OPENQASM 3.0;
    include "stdgates.inc";
    qubit[4] q;
    qubit[1] ancilla;
    x ancilla[0];
    h q[0];
    h q[1];
    h q[2];
    h q[3];
    h ancilla[0];
    cx q[0], ancilla[0];
    cx q[1], ancilla[0];
    cx q[2], ancilla[0];
    cx q[3], ancilla[0];
    h q[0];
    h q[1];
    h q[2];
    h q[3];
    """


@pytest.fixture
def deutch_jozsa_ionq() -> IonQDictType:
    """Return the expected IonQDictType for the DJ algorithm."""
    return {
        "format": InputFormat.CIRCUIT.value,
        "qubits": 5,
        "circuit": [
            {"gate": "x", "target": 0},
            {"gate": "h", "target": 0},
            {"gate": "h", "target": 1},
            {"gate": "h", "target": 2},
            {"gate": "h", "target": 3},
            {"gate": "h", "target": 0},
            {"gate": "cnot", "control": 0, "target": 0},
            {"gate": "cnot", "control": 1, "target": 0},
            {"gate": "cnot", "control": 2, "target": 0},
            {"gate": "cnot", "control": 3, "target": 0},
            {"gate": "h", "target": 0},
            {"gate": "h", "target": 1},
            {"gate": "h", "target": 2},
            {"gate": "h", "target": 3},
        ],
        "gateset": GateSet.QIS.value,
    }


@pytest.fixture
def ionq_native_gates_qasm() -> Qasm3StringType:
    """Return a QASM 3.0 using only IonQ native gates."""
    return """
    OPENQASM 3.0;
    qubit[3] q;
    ms(0,0,0) q[0], q[1];
    ms(-0.5,0.6,0.1) q[1], q[2];
    gpi(0) q[0];
    gpi2(0.2) q[1];
    """


@pytest.fixture
def ionq_native_gates_dict() -> IonQDictType:
    """Return an IonQDictType for a program using native gates."""
    return {
        "format": InputFormat.CIRCUIT.value,
        "gateset": GateSet.NATIVE.value,
        "qubits": 3,
        "circuit": [
            {"gate": "ms", "targets": [0, 1], "phases": [0, 0], "angle": 0.0},
            {"gate": "ms", "targets": [1, 2], "phases": [-0.5, 0.6], "angle": 0.1},
            {"gate": "gpi", "phase": 0, "target": 0},
            {"gate": "gpi2", "phase": 0.2, "target": 1},
        ],
    }


@pytest.mark.skipif(sys.version_info < (3, 11), reason="Requires Python 3.10 or higher")
def test_qasm3_to_ionq_no_pyqasm(deutsch_jozsa_qasm3):
    """Test transpiling the Deutsch-Jozsa algorithm from QASM 3.0 to IonQDictType."""
    with patch(
        "qbraid.transpiler.conversions.qasm3.qasm3_to_ionq.pyqasm.dumps",
        side_effect=Exception("Mocked exception"),
    ):
        with pytest.raises(ProgramConversionError) as excinfo:
            qasm3_to_ionq(deutsch_jozsa_qasm3)
        assert "Failed to parse gate data from OpenQASM string." in str(excinfo.value)


def test_qasm3_to_ionq_deutch_jozsa(
    deutsch_jozsa_qasm3, deutch_jozsa_qasm3_unrolled, deutch_jozsa_ionq
):
    """Test transpiling the Deutsch-Jozsa algorithm from QASM 3.0 to IonQDictType."""
    pyqasm_installed = importlib.util.find_spec("pyqasm") is not None
    qasm_program = deutsch_jozsa_qasm3 if pyqasm_installed else deutch_jozsa_qasm3_unrolled
    ionq_program = qasm3_to_ionq(qasm_program)
    assert ionq_program == deutch_jozsa_ionq


def test_qasm3_to_ionq_deutch_jozsa_pyqasm_mocked(
    deutsch_jozsa_qasm3, deutch_jozsa_qasm3_unrolled, deutch_jozsa_ionq
):
    """Test Deutch-Jozsa conversion with mock pyqasm import and unroll."""
    mock_module = Mock()
    mock_module.unroll = Mock()
    mock_pyqasm = Mock()
    mock_pyqasm.dumps.return_value = deutch_jozsa_qasm3_unrolled

    mock_pyqasm.loads.return_value = mock_module

    with patch.dict("sys.modules", {"pyqasm": mock_pyqasm}):
        qasm_program = deutsch_jozsa_qasm3
        ionq_program = qasm3_to_ionq(qasm_program)
        assert ionq_program == deutch_jozsa_ionq


def test_qasm3_to_ionq_native_gates(ionq_native_gates_qasm, ionq_native_gates_dict):
    """Test transpiling a program using IonQ native gates to IonQDictType."""
    ionq_program = qasm3_to_ionq(ionq_native_gates_qasm)
    assert ionq_program["format"] == ionq_native_gates_dict["format"]
    assert ionq_program["qubits"] == ionq_native_gates_dict["qubits"]
    assert ionq_program["gateset"] == ionq_native_gates_dict["gateset"]
    assert ionq_program["circuit"] == ionq_native_gates_dict["circuit"]
    assert len(ionq_program) == 4


def test_qasm3_to_ionq_zz_native_gate():
    """Test transpiling a program containing the 'zz' native gate to IonQDictType."""
    qasm_program = """
    OPENQASM 3.0;
    qubit[2] q;
    zz(0.12) q[0], q[1];
    """
    expectd_ionq = {
        "format": InputFormat.CIRCUIT.value,
        "gateset": GateSet.NATIVE.value,
        "qubits": 2,
        "circuit": [
            {"gate": "zz", "targets": [0, 1], "angle": 0.12},
        ],
    }

    assert qasm3_to_ionq(qasm_program) == expectd_ionq


@pytest.mark.parametrize(
    "qasm_code, error_message",
    [
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    ms(1.1,0,0) q[0], q[1];
    """,
            "Invalid phase value",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    ms(0,-1.5,0) q[0], q[1];
    """,
            "Invalid phase value",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    gpi(-6) q[0], q[1];
    """,
            "Invalid phase value",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    gpi2(3.0) q[0], q[1];
    """,
            "Invalid phase value",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    ms(0,0,0.26) q[0], q[1];
    """,
            "Invalid angle value",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    ms(0,0,-0.1) q[0], q[1];
    """,
            "Invalid angle value",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    zz(abc) q[0], q[1];
    """,
            "Invalid angle value 'abc'",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    gpi(0) q[1];
    zz(2*pi) q[0], q[1];
    """,
            "Invalid angle value '2 * pi'",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    cgpi2(0) q[0], q[1];
    zz(2*pi) q[0], q[1];
    """,
            "Invalid angle value '2 * pi'",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    zz(0.15) q[0], q[1];
    zz(3.0) q[0], q[1];
    """,
            "Invalid angle value '3.0'",
        ),
    ],
)
def test_qasm3_to_ionq_invalid_params(qasm_code, error_message):
    """Test that qasm3_to_ionq raises an error when the circuit contains invalid parameters."""
    with pytest.raises(ProgramConversionError) as excinfo:
        qasm3_to_ionq(qasm_code)
    assert error_message in str(excinfo.value)


@pytest.mark.parametrize(
    "qasm_code, ionq_dict",
    [
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    zz(0.45) q[0], q[1];
    """,
            {
                "qubits": 2,
                "circuit": [{"gate": "zz", "rotation": 0.45, "targets": [0, 1]}],
                "gateset": "qis",
                "format": "ionq.circuit.v0",
            },
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    zz(-0.9) q[0], q[1];
    """,
            {
                "qubits": 2,
                "circuit": [{"gate": "zz", "rotation": -0.9, "targets": [0, 1]}],
                "gateset": "qis",
                "format": "ionq.circuit.v0",
            },
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    zz(0.1) q[0], q[1];
    """,
            {
                "qubits": 2,
                "circuit": [{"gate": "zz", "angle": 0.1, "targets": [0, 1]}],
                "gateset": "native",
                "format": "ionq.circuit.v0",
            },
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    h q[0];
    zz(pi/16) q[0], q[1];
    """,
            {
                "qubits": 2,
                "circuit": [
                    {"gate": "h", "target": 0},
                    {"gate": "zz", "rotation": 0.19634954084936207, "targets": [0, 1]},
                ],
                "gateset": "qis",
                "format": "ionq.circuit.v0",
            },
        ),
    ],
)
def test_qasm3_to_ionq_zz_context(qasm_code, ionq_dict):
    """Test that qasm3_to_ionq correctly labels the gate parameter based on the value."""
    assert qasm3_to_ionq(qasm_code) == ionq_dict


@pytest.mark.parametrize(
    "qasm_code, error_message",
    [
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    gpi q[0];
    """,
            "Phase parameter is required",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[1] q;
    rz q[0];
    """,
            "Rotation parameter is required",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    crz q[0], q[1];
    """,
            "Rotation parameter is required",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    zz q[0], q[1];
    """,
            "Angle parameter is required",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[1] q;
    invalid_gate q[0];
    """,
            "Gate 'invalid_gate' not supported",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[3] q;
    ch q[0], q[1], q[2];
    """,
            "Invalid number of qubits",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[3] q;
    cnot q[0], q[1], q[2];
    """,
            "Invalid number of qubits",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[4] q;
    ccnot q[0], q[1], q[2], q[3];
    """,
            "Invalid number of qubits",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[3] q;
    cgpi2(0) q[0], q[1], q[3];
    """,
            "Invalid number of qubits",
        ),
        (
            """
    OPENQASM 3.0;
    qubit[2] q;
    cgpi q[0], q[1];
    """,
            "Phase parameter is required",
        ),
    ],
)
def test_openqasm3_to_ionq_value_errors(qasm_code, error_message):
    """Test that openqasm3_to_ionq raises an error when the circuit contains
    a gate that is missing required parameters or is not supported."""
    with pytest.raises(ValueError) as excinfo:
        openqasm3_to_ionq(qasm_code)
    assert error_message in str(excinfo.value)


def test_qasm3_to_ionq_mixed_gate_types_raises_value_error():
    """Test that qasm3_to_ionq raises an error when the circuit contains mixed gate types."""
    mixed_gate_qasm = """
    OPENQASM 3.0;
    qubit[2] q;
    h q[0];
    h q[1];
    gpi(0) q[0], q[1];
    """
    with pytest.raises(ProgramConversionError) as excinfo:
        qasm3_to_ionq(mixed_gate_qasm)
    assert "Cannot mix native and QIS gates in the same circuit." in str(excinfo.value)


def test_extract_params_index_error_caught():
    """Test that the extract_params returns empty list for non-parametric gates."""
    h_gate_qasm = """
    OPENQASM 3.0;
    qubit[1] q;
    h q[0];    
    """
    program = parse(h_gate_qasm)
    statement = program.statements[1]
    assert isinstance(statement, openqasm3.ast.QuantumGate)
    assert extract_params(statement) == []


@pytest.mark.skip(reason="To validate in pyqasm through definition of ms gate")
@pytest.mark.parametrize(
    "program_text",
    [
        """
    OPENQASM 3.0;
    qubit[2] q;
    ms(0,0,0,0) q[0], q[1];
    """,
        """
    OPENQASM 3.0;
    qubit[2] q;
    ms q[0], q[1];
    """,
    ],
)
def test_ionq_ms_gate_wrong_number_params(program_text):
    """Test ValueError is raised when 'ms' gate has an invalid number of parameters."""
    program = OpenQasm3Program(program_text)

    with pytest.raises(ValueError) as excinfo:
        _ = _parse_gates(program)
    assert "Invalid number of parameters" in str(excinfo.value)


@pytest.fixture
def controlled_gates_qasm() -> Qasm3StringType:
    """Return a QASM 3.0 string containing various controlled gates."""
    return """
    OPENQASM 3.0;
    include "stdgates.inc";
    qubit[5] q;

    cx q[0], q[4];
    cy q[1], q[4];
    cz q[2], q[4];
    crx(0.7853981633974483) q[3], q[4];
    cry(1.5707963267948966) q[3], q[4];
    crz(2.356194490192345) q[3], q[4];
    ch q[0], q[4];
    ccx q[0], q[1], q[4];
    """


@pytest.fixture
def controlled_gates_ionq() -> IonQDictType:
    """Return the expected IonQDictType for controlled gated circuit."""
    return {
        "format": InputFormat.CIRCUIT.value,
        "qubits": 5,
        "circuit": [
            {"gate": "cnot", "control": 0, "target": 4},
            {"gate": "y", "control": 1, "target": 4},
            {"gate": "z", "control": 2, "target": 4},
            {"gate": "rx", "control": 3, "target": 4, "rotation": 0.7853981633974483},
            {"gate": "ry", "control": 3, "target": 4, "rotation": 1.5707963267948966},
            {"gate": "rz", "control": 3, "target": 4, "rotation": 2.356194490192345},
            {"gate": "h", "control": 0, "target": 4},
            {"gate": "cnot", "controls": [0, 1], "target": 4},
        ],
        "gateset": GateSet.QIS.value,
    }


def test_qasm3_to_ionq_controlled_gates(controlled_gates_qasm, controlled_gates_ionq):
    """Test transpiling QASM 3.0 program containing various controlled gates to IonQDictType."""
    ionq_program = qasm3_to_ionq(controlled_gates_qasm)
    assert ionq_program == controlled_gates_ionq


@pytest.fixture
def controlled_gates_native_qasm() -> Qasm3StringType:
    """Return a QASM 3.0 string containing native controlled gates."""
    return """
    OPENQASM 3.0;
    include "stdgates.inc";
    // this declaration is required so that the pyqasm parser can correctly
    // identify the gate
    gate cgpi2(a) q1, q2 {
    }
    qubit[2] q;

    cgpi2(0.2) q[0], q[1];
    """


@pytest.fixture
def controlled_gates_native_ionq() -> IonQDictType:
    """Return the expected IonQDictType for circuit with native controlled gates."""
    return {
        "format": InputFormat.CIRCUIT.value,
        "qubits": 2,
        "circuit": [
            {"gate": "gpi2", "phase": 0.2, "control": 0, "target": 1},
        ],
        "gateset": GateSet.NATIVE.value,
    }


def test_qasm3_to_ionq_native_controlled_gates(
    controlled_gates_native_qasm, controlled_gates_native_ionq
):
    """Test transpiling QASM 3.0 program containing native controlled gates to IonQDictType."""
    ionq_program = qasm3_to_ionq(controlled_gates_native_qasm)
    assert ionq_program == controlled_gates_native_ionq



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Cirq conversions

.. currentmodule:: qbraid.transpiler.conversions.cirq

Functions
----------

.. autosummary::
   :toctree: ../stubs/

   cirq_to_braket
   cirq_to_qasm2
   cirq_to_pyquil
   cirq_to_pyqir
   cirq_to_stim
   stim_to_cirq

"""
from .cirq_extras import cirq_to_pyqir, cirq_to_stim, stim_to_cirq
from .cirq_to_braket import cirq_to_braket
from .cirq_to_pyquil import cirq_to_pyquil
from .cirq_to_qasm2 import cirq_to_qasm2

__all__ = [
    "cirq_to_braket",
    "cirq_to_qasm2",
    "cirq_to_pyquil",
    "cirq_to_pyqir",
    "cirq_to_stim",
    "stim_to_cirq",
]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for QASM preprocessing functions

"""

import pyqasm
import pytest
from qiskit import QuantumCircuit

from qbraid.interface import circuits_allclose
from qbraid.programs import load_program
from qbraid.transpiler.conversions.qasm2 import qasm2_to_cirq

qasm_0 = """OPENQASM 2.0;
include "qelib1.inc";
gate rzx(param0) q0,q1 { h q1; cx q0,q1; rz(-pi/4) q1; cx q0,q1; h q1; }
gate ecr q0,q1 { rzx(pi/4) q0,q1; x q0; rzx(-pi/4) q0,q1; }
gate rzx_6320157840(param0) q0,q1 { h q1; cx q0,q1; rz(2.3200048200765524) q1; cx q0,q1; h q1; }
qreg q[4];
cry(5.518945082555831) q[0],q[1];
u(5.75740842861076,5.870881397684582,1.8535618384181967) q[2];
ecr q[3],q[0];
rzx_6320157840(2.3200048200765524) q[2],q[1];
rccx q[1],q[2],q[3];
csx q[0],q[1];
rzx(5.603791034636421) q[2],q[0];
"""

qasm_1 = """
OPENQASM 2.0;
include "qelib1.inc";
qreg q[5];
rccx q[1],q[2],q[0];
cu(5.64,3.60,3.73, 5.68) q[1],q[0];
c3x q[1],q[3],q[0],q[2];
c3sqrtx q[3],q[1],q[2],q[0];
c4x q[2],q[0],q[1],q[4],q[3];
rc3x q[1],q[2],q[0],q[3];
"""

qasm_lst = [qasm_0, qasm_1]


def strings_equal(s1, s2):
    """Check if two strings are equal, ignoring spaces and newlines."""
    s1_clean = s1.replace(" ", "").replace("\n", "")
    s2_clean = s2.replace(" ", "").replace("\n", "")
    return s1_clean == s2_clean


def _test_qasm_preprocess(qasm_str):
    """Test converting qasm string to format supported by Cirq parser"""
    qiskit_circuit = QuantumCircuit().from_qasm_str(qasm_str)
    qasm_module = pyqasm.loads(qasm_str)
    qasm_module.unroll()
    supported_qasm = pyqasm.dumps(qasm_module)
    cirq_circuit = qasm2_to_cirq(supported_qasm)
    qprogram = load_program(cirq_circuit)
    qprogram._convert_to_line_qubits()
    cirq_circuit_compat = qprogram.program
    assert circuits_allclose(cirq_circuit_compat, qiskit_circuit)


def test_preprocess_qasm_0():
    """Test converting qasm string to format supported by Cirq parser"""
    _test_qasm_preprocess(qasm_0)


@pytest.mark.skip(reason="Mapping not implemented yet for complex quantum gates")
def test_preprocess_qasm_1():
    """Test converting qasm string to format supported by Cirq parser"""
    _test_qasm_preprocess(qasm_1)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for managing quantum program type aliases.

"""
from unittest.mock import Mock

import pytest

from qbraid.programs._import import _assign_default_type_alias, _get_class
from qbraid.programs.alias_manager import (
    _get_program_type_alias,
    find_str_type_alias,
    get_program_type_alias,
    get_qasm_type_alias,
)
from qbraid.programs.exceptions import ProgramTypeError
from qbraid.programs.exceptions import QasmError as QbraidQasmError
from qbraid.programs.registry import derive_program_type_alias

from ..fixtures import packages_bell

QASM_BELL_DATA = [
    (
        """
OPENQASM 2.0;
include "qelib1.inc";
qreg q[2];
h q[0];
cx q[0],q[1];
        """,
        "qasm2",
    ),
    (
        """
OPENQASM 3;
include "stdgates.inc";
qubit[2] q;
h q[0];
cx q[0], q[1];
        """,
        "qasm3",
    ),
    (
        """
OPENQASM 3.0;
bit[2] __bits__;
qubit[2] __qubits__;
h __qubits__[0];
cnot __qubits__[0], __qubits__[1];
__bits__[0] = measure __qubits__[0];
__bits__[1] = measure __qubits__[1];
        """,
        "qasm3",
    ),
]

QASM_ERROR_DATA = [
    """
OPENQASM xyz;
include "qelib1.inc";
qreg q[2];
creg c[2];
h q[0;
cx q[0], q[1];
measure q[0] -> c[0];
measure q[1] -> c[1];

        """,
    """
OPENQASM #@!$;
include "stdgates.inc";
bit c[2];
qubit q[2];
h q[0]
cx q[0], q[1];
measure q -> c;
        """,
]


@pytest.mark.parametrize("qasm_str, expected_version", QASM_BELL_DATA)
def test_get_qasm_type_alias(qasm_str: str, expected_version: str):
    """Test getting QASM version"""
    alias = get_qasm_type_alias(qasm_str)
    assert str(alias) == str(expected_version)


@pytest.mark.parametrize("qasm_str", QASM_ERROR_DATA)
def test_get_qasm_type_alias_error(qasm_str):
    """Test getting QASM version"""
    with pytest.raises(QbraidQasmError):
        get_qasm_type_alias(qasm_str)


@pytest.mark.parametrize("bell_circuit", packages_bell, indirect=True)
def test_get_program_type_alias(bell_circuit):
    """Test that the correct package is returned for a given program."""
    circuit, expected_package = bell_circuit
    package_name = get_program_type_alias(circuit)
    assert package_name == expected_package


@pytest.mark.parametrize("program,expected_package", [(Mock(), "unittest")])
def test_get_program_type_alias_required_supported_false(program, expected_package):
    """Test that None or module name is returned for unsupported package when
    require supported is given as False."""
    package = derive_program_type_alias(program)
    assert package == expected_package


def test_raise_error_unuspported_source_program():
    """Test that an error is raised if source program is not supported."""
    with pytest.raises(ProgramTypeError):
        get_program_type_alias(Mock())


@pytest.mark.parametrize(
    "item",
    ["OPENQASM 1.0; bad operation", "OPENQASM -3.1; bad operation", "DECLARE ro BIT[1]", "circuit"],
)
def test_bad_source_openqasm_program(item):
    """Test raising ProgramTypeError converting invalid OpenQASM program string"""
    with pytest.raises(ProgramTypeError):
        get_program_type_alias(item)


@pytest.mark.parametrize("item", [1, None])
def test_bad_source_program_type(item):
    """Test raising ProgramTypeError converting circuit of non-supported type"""
    with pytest.raises(ProgramTypeError):
        get_program_type_alias(item)


def test_get_program_type_alias_with_type_instead_of_instance():
    """Test error raised when a type rather than an instance is provided."""
    with pytest.raises(ProgramTypeError) as excinfo:
        _get_program_type_alias(int)
    assert "Expected an instance of a quantum program, not a type." in str(excinfo.value)


def test_get_program_type_alias_with_string_returning_package(monkeypatch):
    """Test that a valid string type alias is correctly returned."""
    monkeypatch.setattr(
        "qbraid.programs.alias_manager.find_str_type_alias", lambda: "valid_package"
    )
    monkeypatch.setattr(
        "qbraid.programs.alias_manager.get_qasm_type_alias",
        lambda x: (_ for _ in ()).throw(QbraidQasmError("error")),
    )

    result = _get_program_type_alias("some qasm program string")
    assert (
        result == "valid_package"
    ), "Did not return expected package when valid package is available"


def test_get_program_type_alias_with_multiple_matches(monkeypatch):
    """Test error raised when program matches multiple registered types."""
    monkeypatch.setattr(
        "qbraid.programs.alias_manager.QPROGRAM_REGISTRY", {"type1": Mock, "type2": Mock}
    )
    with pytest.raises(ProgramTypeError) as excinfo:
        _get_program_type_alias(Mock())
    assert "matches multiple registered program types" in str(excinfo.value)


def test_get_program_type_alias_safe_flag_handling():
    """Test handling of 'safe' flag with non-matching program types."""
    result = get_program_type_alias(object(), safe=True)
    assert result is None, "Should return None when safe is True and no matching types are found"


def test_find_single_str_type_alias():
    """Test the case where there is exactly one valid string key."""
    registry = {
        "qasm2": int,
        "qasm3": float,
        "custom_alias": str,
    }
    assert find_str_type_alias(registry) == "custom_alias"


def test_find_str_type_alias_raises_value_error():
    """Test the case where there are multiple valid string keys, expecting a ValueError."""
    registry = {
        "qasm2": int,
        "qasm3": float,
        "alias1": str,
        "alias2": str,
    }
    with pytest.raises(ValueError) as excinfo:
        find_str_type_alias(registry)
    assert "Multiple additional keys with type 'str' found: ['alias1', 'alias2']" in str(
        excinfo.value
    )


def test_find_no_str_type_alias():
    """Test the case where there are no valid string keys."""
    registry = {
        "qasm2": int,
        "qasm3": float,
    }
    assert find_str_type_alias(registry) is None


class SinglePartModule:
    """Test class with a single-part module name."""

    __module__ = "quantum"


class MultiPartModule:
    """Test class with a multi-part module name."""

    __module__ = "parent.child"


class MainModule:
    """Test class with the __main__ module."""

    __module__ = "__main__"


class BuiltinsModule:
    """Test class with the builtins module."""

    __module__ = "builtins"


def test_submodule_multi_part():
    """Check alias creation for multi-part module with submodule flag."""
    assert derive_program_type_alias(MultiPartModule, use_submodule=True) == "parent_child"


def test_submodule_single_part_error():
    """Ensure ValueError for single-part module with submodule flag."""
    with pytest.raises(
        ValueError, match="Module name does not have at least two parts for submodule aliasing."
    ):
        derive_program_type_alias(SinglePartModule, use_submodule=True)


def test_submodule_main_error():
    """Ensure ValueError for __main__ module with submodule flag."""
    with pytest.raises(
        ValueError, match="Cannot use submodule aliasing with __main__ or builtins module."
    ):
        derive_program_type_alias(MainModule, use_submodule=True)


def test_submodule_builtins_error():
    """Ensure ValueError for builtins module with submodule flag."""
    with pytest.raises(
        ValueError, match="Cannot use submodule aliasing with __main__ or builtins module."
    ):
        derive_program_type_alias(BuiltinsModule, use_submodule=True)


def test_submodule_insufficient_error():
    """Ensure ValueError for insufficient module parts with submodule flag."""
    with pytest.raises(
        ValueError, match="Module name does not have at least two parts for submodule aliasing."
    ):
        derive_program_type_alias(SinglePartModule, use_submodule=True)


def test_single_part_module_name():
    """Test alias assignment with a single-part module name."""
    imported_aliases = {"quantum": object}
    alias = _assign_default_type_alias(imported_aliases, SinglePartModule)
    assert alias == "quantum_singlepartmodule"


def test_duplicate_alias_raises_value_error():
    """Ensure ValueError is raised for duplicate alias."""
    imported_aliases = {"quantum": object, "quantum_singlepartmodule": object}
    with pytest.raises(ValueError, match="Duplicate alias quantum_singlepartmodule"):
        _assign_default_type_alias(imported_aliases, SinglePartModule)


def test_get_ionq_program_type_alias():
    """Test getting the IonQ program type alias."""
    circuit = {
        "qubits": 3,
        "circuit": [
            {"gate": "h", "target": 0},
            {"gate": "cnot", "control": 0, "target": 1},
            {"gate": "cnot", "control": 0, "target": 2},
        ],
    }
    assert _get_program_type_alias(circuit) == "ionq"


def test_get_class_none_for_unsupported_program():
    """Test that None is returned for unsupported program type module."""
    with pytest.raises(ValueError) as excinfo:
        _get_class("not a module")
    assert "Unsupported module 'not a module'" == str(excinfo.value)



# Copyright (C) 2025 qBraid
# Copyright (C) 2025 Microsoft Corporation
#
# This file is part of the qBraid-SDK.
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. This specific file, adapted from Microsoft, is dual-licensed under both
# the MIT License, and the GPL v3. You may not use this file except in compliance
# with the applicable license. You may obtain a copy of the MIT License at
#
# https://opensource.org/license/mit
#
# This file includes code adapted from Microsoft (https://github.com/microsoft/azure-quantum-python)
# with modifications by qBraid. The original copyright notice is included above.
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# qbraid: skip-header

"""
Module defining AzureResultBuilder class

"""
import ast
import datetime
import json
import os
import re
from typing import Any, Optional, Union

import numpy as np
from azure.quantum import Job

from qbraid.runtime.ionq.job import IonQJob
from qbraid.runtime.postprocess import counts_to_probabilities, normalize_data

from .io_format import OutputDataFormat


class AzureResultBuilder:
    """Class to format Azure Quantum job results."""

    def __init__(self, azure_job: Job):
        self._azure_job = azure_job

    @property
    def job(self) -> Job:
        """Return the Azure Quantum job."""
        return self._azure_job

    @property
    def from_simulator(self) -> bool:
        """Return whether the job was executed on a simulator."""
        return self._azure_job.details.target.split(".", 2)[1] != "qpu"

    def _shots_count(self) -> Optional[int]:
        """Return the number of shots used in the job."""
        # Some providers use 'count', some other 'shots', give preference to 'count':
        input_params = self.job.details.input_params
        return input_params.get("count", input_params.get("shots"))

    def _format_results(
        self, sampler_seed: Optional[int] = None
    ) -> Union[list[dict[str, Any]], dict[str, Any]]:
        """
        Populates the results datastructures in a format that
        is compatible with qBraid runtime.

        """
        if self.job.details.output_data_format == OutputDataFormat.MICROSOFT_V2.value:
            return self._format_microsoft_v2_results()

        success = self.job.details.status == "Succeeded"

        job_result = {
            "data": {},
            "success": success,
            "header": {},
        }

        if success:
            if self.job.details.output_data_format == OutputDataFormat.MICROSOFT_V1.value:
                job_result["data"] = self._format_microsoft_results(sampler_seed=sampler_seed)
            elif self.job.details.output_data_format == OutputDataFormat.IONQ.value:
                job_result["data"] = self._format_ionq_results()
            elif self.job.details.output_data_format == OutputDataFormat.QUANTINUUM.value:
                job_result["data"] = self._format_quantinuum_results()
            elif self.job.details.output_data_format == OutputDataFormat.RIGETTI.value:
                job_result["data"] = self._format_rigetti_results()
            elif self.job.details.output_data_format == OutputDataFormat.PASQAL.value:
                job_result["data"] = self._format_analog_results()
            else:
                job_result["data"] = self._format_unknown_results()

        job_result["header"] = self.job.details.metadata or {}
        if "metadata" in job_result["header"]:
            job_result["header"]["metadata"] = json.loads(job_result["header"]["metadata"])

        job_result["shots"] = self._shots_count()
        return job_result

    @staticmethod
    def _draw_random_sample(
        probabilities: dict[str, int], shots: int, sampler_seed: Optional[int] = None
    ) -> dict:
        """Draw a random sample from the given probabilities."""
        norm = sum(probabilities.values())

        if norm != 1:
            if not np.isclose(norm, 1.0, rtol=1e-4):
                raise ValueError(f"Probabilities do not add up to 1: {probabilities}")
            probabilities = {k: v / norm for k, v in probabilities.items()}
        if not sampler_seed:
            current_microtime = int(datetime.datetime.now().timestamp() * 1_000_000)
            random_bytes = os.urandom(4) + current_microtime.to_bytes(8, "little")
            sampler_seed = int.from_bytes(random_bytes, "little") % (2**32 - 1)

        rng = np.random.default_rng(sampler_seed)
        rand_values = rng.choice(list(probabilities.keys()), shots, p=list(probabilities.values()))
        return dict(zip(*np.unique(rand_values, return_counts=True)))

    def _format_ionq_results(self) -> dict[str, Any]:
        """
        Translate IonQ's histogram data into a format that
        can be consumed by qBraid runtime.

        """
        shots = self._shots_count()
        az_result = self.job.get_results()

        if "histogram" not in az_result:
            raise ValueError("Histogram missing from IonQ Job results")

        data = {
            "shots": shots,
            "probabilities": az_result["histogram"],
        }

        raw_counts = IonQJob._get_counts(data)
        counts = normalize_data(raw_counts)
        probabilities = counts_to_probabilities(counts)

        return {"counts": counts, "probabilities": probabilities}

    @staticmethod
    def _qir_to_qbraid_bitstring(obj) -> str:
        """Convert the data structure from Azure into the "schema" used by qBraid."""
        if isinstance(obj, str) and not re.match(r"[\d\s]+$", obj):
            obj = ast.literal_eval(obj)

        if isinstance(obj, tuple):
            # the outermost implied container is a tuple, and each item is
            # associated with a classical register. Azure and qBraid order the
            # registers in opposite directions, so reverse here to match.
            return " ".join(
                [AzureResultBuilder._qir_to_qbraid_bitstring(term) for term in reversed(obj)]
            )
        if isinstance(obj, list):
            # a list is for an individual classical register
            return "".join([str(bit) for bit in obj])

        return str(obj)

    def _format_microsoft_results(self, sampler_seed: Optional[int] = None) -> dict[str, Any]:
        """
        Translate Microsoft's job results histogram into a format that
        can be consumed by qBraid runtime.

        """
        histogram = self.job.get_results()
        shots = self._shots_count()

        counts = {}
        probabilities = {}

        for key in histogram.keys():
            bitstring = self._qir_to_qbraid_bitstring(key)

            value = histogram[key]
            probabilities[bitstring] = value

        if self.from_simulator:
            counts = self._draw_random_sample(probabilities, shots, sampler_seed)
        else:
            counts = {
                bitstring: np.round(shots * value) for bitstring, value in probabilities.items()
            }

        return {"counts": counts, "probabilities": probabilities}

    def _format_quantinuum_results(self) -> dict[str, Any]:
        """
        Translate Quantinuum's histogram data into a format that
        can be consumed by qBraid runtime.

        """
        az_result = self.job.get_results()
        all_bitstrings = [
            bitstrings
            for classical_register, bitstrings in az_result.items()
            if classical_register != "access_token"
        ]
        counts = {}
        combined_bitstrings = ["".join(bitstrings) for bitstrings in zip(*all_bitstrings)]
        shots = len(combined_bitstrings)

        for bitstring in set(combined_bitstrings):
            counts[bitstring] = combined_bitstrings.count(bitstring)

        histogram = {bitstring: count / shots for bitstring, count in counts.items()}

        return {"counts": counts, "probabilities": histogram}

    def _format_rigetti_results(self) -> dict[str, Any]:
        """
        Translate Rigetti's readout data into a format that
        can be consumed by qBraid runtime.

        """
        az_result = self.job.get_results()
        readout = az_result["ro"]
        measurements = ["".join(map(str, row)) for row in readout]
        counts = {row: measurements.count(row) for row in set(measurements)}
        total_counts = sum(counts.values())
        probabilities = {outcome: count / total_counts for outcome, count in counts.items()}
        return {"counts": counts, "probabilities": probabilities}

    def _format_analog_results(self) -> dict[str, Any]:
        """
        Translate Microsoft's AHS job results histogram into a format that
        can be consumed by qBraid runtime.

        """
        histogram = self.job.get_results()
        counts = normalize_data(histogram)
        probabilities = counts_to_probabilities(counts)
        return {"counts": histogram, "probabilities": probabilities}

    def _format_unknown_results(self):
        """Format Job results data when the job output is in an unknown format."""
        return self.job.get_results()

    def _translate_microsoft_v2_results(self) -> list[tuple[int, dict[str, Any]]]:
        """
        Translate Microsoft's batching job results histograms into a format that
        can be consumed by qBraid runtime.

        """
        az_result = self.job.get_results()

        if "DataFormat" not in az_result:
            raise ValueError("DataFormat missing from Job results")

        if "Results" not in az_result:
            raise ValueError("Results missing from Job results")

        histograms = []
        results = az_result["Results"]
        for circuit_results in results:
            counts = {}
            probabilities = {}

            if "TotalCount" not in circuit_results:
                raise ValueError("TotalCount missing from Job results")

            total_count: int = circuit_results["TotalCount"]

            if total_count <= 0:
                raise ValueError("TotalCount must be a positive non-zero integer")

            if "Histogram" not in circuit_results:
                raise ValueError("Histogram missing from Job results")

            histogram = circuit_results["Histogram"]
            for result in histogram:
                if "Display" not in result:
                    raise ValueError("Dispaly missing from histogram result")

                if "Count" not in result:
                    raise ValueError("Count missing from histogram result")

                bitstring = self._qir_to_qbraid_bitstring(result["Display"])
                count = result["Count"]
                probability = count / total_count
                counts[bitstring] = count
                probabilities[bitstring] = probability
            histograms.append((total_count, {"counts": counts, "probabilities": probabilities}))
        return histograms

    def _get_entry_point_names(self) -> list[str]:
        """Get QIR entry point names from the input parameters."""
        input_params = self.job.details.input_params
        # All V2 output is a list of entry points
        entry_points = input_params["items"]
        entry_point_names = []
        for entry_point in entry_points:
            if "entryPoint" not in entry_point:
                raise ValueError("Entry point input_param is missing an 'entryPoint' field")
            entry_point_names.append(entry_point["entryPoint"])
        return entry_point_names if len(entry_point_names) > 0 else ["main"]

    def _format_microsoft_v2_results(self) -> list[dict[str, Any]]:
        """
        Translate Microsoft's batching job results histograms into a format that
        can be consumed by qBraid runtime.

        """
        status = self.job.details.status
        success = status == "Succeeded"

        if not success:
            return [
                {
                    "data": {},
                    "success": False,
                    "header": {},
                    "shots": 0,
                }
            ]

        entry_point_names = self._get_entry_point_names()

        results = self._translate_microsoft_v2_results()

        if len(results) != len(entry_point_names):
            raise ValueError(
                "The number of experiment results does not match the number of experiment names"
            )

        return [
            {
                "data": result,
                "success": success,
                "shots": total_count,
                "name": name,
                "status": status,
                "header": {"name": name},
            }
            for name, (total_count, result) in zip(entry_point_names, results)
        ]

    def get_results(
        self, timeout: Optional[int] = None, sampler_seed: Optional[int] = None
    ) -> list[dict[str, Any]]:
        """Return the results of the job."""
        self.job.wait_until_completed(timeout_secs=timeout)

        results = self._format_results(sampler_seed=sampler_seed)
        results = results if isinstance(results, list) else [results]
        return results

    def get_counts(self) -> Union[dict[str, int], list[dict[str, int]]]:
        """Return the raw counts from the result data."""
        results = self.get_results()

        if len(results) == 1:
            return results[0]["data"]["counts"] if results[0]["success"] else {}

        return [result["data"]["counts"] if result["success"] else {} for result in results]
